"""Base class for generating an S matrix automatically from tidy3d simulations and port definitions."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, Optional, Union, get_args

import pydantic.v1 as pd

from tidy3d.components.autograd.constants import MAX_NUM_ADJOINT_PER_FWD
from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.geometry.utils import _shift_value_signed
from tidy3d.components.simulation import Simulation
from tidy3d.components.types import Complex, FreqArray
from tidy3d.components.validators import (
    assert_unique_names,
    validate_freqs_min,
    validate_freqs_not_empty,
    validate_freqs_unique,
)
from tidy3d.constants import HERTZ
from tidy3d.exceptions import SetupError, Tidy3dKeyError
from tidy3d.log import log
from tidy3d.plugins.smatrix.ports.modal import Port
from tidy3d.plugins.smatrix.ports.types import TerminalPortType
from tidy3d.plugins.smatrix.ports.wave import WavePort
from tidy3d.plugins.smatrix.types import Element, MatrixIndex, NetworkElement, NetworkIndex

if TYPE_CHECKING:
    from tidy3d.web.core.types import PayType

# fwidth of gaussian pulse in units of central frequency
FWIDTH_FRAC = 1.0 / 10
DEFAULT_DATA_DIR = "."

IndexType = Union[MatrixIndex, NetworkIndex]
ElementType = Union[Element, NetworkElement]
TaskNameFormat = Literal["RF", "PF"]


class AbstractComponentModeler(ABC, Tidy3dBaseModel):
    """Tool for modeling devices and computing port parameters."""

    name: str = pd.Field(
        "",
        title="Name",
    )

    simulation: Simulation = pd.Field(
        ...,
        title="Simulation",
        description="Simulation describing the device without any sources present.",
    )

    ports: tuple[Union[Port, TerminalPortType], ...] = pd.Field(
        (),
        title="Ports",
        description="Collection of ports describing the scattering matrix elements. "
        "For each input mode, one simulation will be run with a modal source.",
    )

    freqs: FreqArray = pd.Field(
        ...,
        title="Frequencies",
        description="Array or list of frequencies at which to compute port parameters.",
        units=HERTZ,
    )

    remove_dc_component: bool = pd.Field(
        True,
        title="Remove DC Component",
        description="Whether to remove the DC component in the Gaussian pulse spectrum. "
        "If ``True``, the Gaussian pulse is modified at low frequencies to zero out the "
        "DC component, which is usually desirable so that the fields will decay. However, "
        "for broadband simulations, it may be better to have non-vanishing source power "
        "near zero frequency. Setting this to ``False`` results in an unmodified Gaussian "
        "pulse spectrum which can have a nonzero DC component.",
    )

    run_only: Optional[tuple[IndexType, ...]] = pd.Field(
        None,
        title="Run Only",
        description="Set of matrix indices that define the simulations to run. "
        "If ``None``, simulations will be run for all indices in the scattering matrix. "
        "If a tuple is given, simulations will be run only for the given matrix indices.",
    )

    element_mappings: tuple[tuple[ElementType, ElementType, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Tuple of S matrix element mappings, each described by a tuple of "
        "(input_element, output_element, coefficient), where the coefficient is the "
        "element_mapping coefficient describing the relationship between the input and output "
        "matrix element. If all elements of a given column of the scattering matrix are defined "
        "by ``element_mappings``, the simulation corresponding to this column is skipped automatically.",
    )

    @pd.validator("simulation", always=True)
    def _sim_has_no_sources(cls, val):
        """Make sure simulation has no sources as they interfere with tool."""
        if len(val.sources) > 0:
            raise SetupError(f"'{cls.__name__}.simulation' must not have any sources.")
        return val

    @pd.validator("ports", always=True)
    def _warn_rf_license(cls, val):
        """Warn about new licensing requirements for RF ports."""
        rf_port = False
        TerminalPortTypeTuple = get_args(TerminalPortType)
        for port in val:
            if type(port) in TerminalPortTypeTuple:
                rf_port = True
                break
        if rf_port:
            log.warning(
                "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
                log_once=True,
            )
        return val

    @pd.validator("element_mappings", always=True)
    def _validate_element_mappings(cls, element_mappings, values):
        """
        Validate that each source index referenced in element_mappings is included in run_only.
        """
        run_only = values.get("run_only")
        if run_only is None:
            return element_mappings

        valid_set = set(run_only)
        invalid_indices = set()
        for mapping in element_mappings:
            input_element = mapping[0]
            output_element = mapping[1]
            for source_index in [input_element[1], output_element[1]]:
                if source_index not in valid_set:
                    invalid_indices.add(source_index)
        if invalid_indices:
            raise SetupError(
                f"'element_mappings' references source index(es) {invalid_indices} "
                f"that are not present in run_only: {run_only}."
            )
        return element_mappings

    _freqs_not_empty = validate_freqs_not_empty()
    _freqs_lower_bound = validate_freqs_min()
    _freqs_unique = validate_freqs_unique()

    @staticmethod
    def get_task_name(
        port: Port, mode_index: Optional[int] = None, format: Optional[TaskNameFormat] = "RF"
    ) -> str:
        """Generates a standardized task name from a port object.

        This method creates a unique string identifier for a simulation task based on
        a port. The naming convention can be controlled by specifying a mode index
        directly, which takes precedence, or by selecting a predefined format.

        Parameters
        ----------
        port : Port
            The port object from which to derive the base name.
        mode_index : Optional[int], optional
            If provided, this index is appended to the port name (e.g., 'port_1@1'),
            overriding the `format` argument. Defaults to None.
        format : TaskNameFormat, optional
            Specifies the naming convention to use when `mode_index` is not provided.
            - "RF": Returns the plain port name (e.g., "port_1").
            - "PF": Appends a zero index to the name (e.g., "port_1@0").
            Defaults to "RF".

        Returns
        -------
        str
            The formatted task name string.

        Raises
        ------
        ValueError
            If an invalid `format` string is provided.
        """
        if mode_index is not None:
            return f"{port.name}@{mode_index}"
        if format == "PF":
            port_name = f"{port.name}@0"
        elif format == "RF":
            port_name = f"{port.name}"
        else:
            raise ValueError(
                f"Format '{format}' invalid for port-name task creation. Must be defined in {TaskNameFormat}"
            )
        return port_name

    def get_port_by_name(self, port_name: str) -> Port:
        """Get the port from the name."""
        ports = [port for port in self.ports if port.name == port_name]
        if len(ports) == 0:
            raise Tidy3dKeyError(f'Port "{port_name}" not found.')
        return ports[0]

    @property
    @abstractmethod
    def matrix_indices_monitor(self) -> tuple[IndexType, ...]:
        """Abstract property for all matrix indices that will be used to collect data."""

    @cached_property
    def matrix_indices_source(self) -> tuple[IndexType, ...]:
        """Tuple of all the source matrix indices, which may be less than the total number of
        ports."""
        if self.run_only is not None:
            return self.run_only
        return self.matrix_indices_monitor

    @cached_property
    def matrix_indices_run_sim(self) -> tuple[IndexType, ...]:
        """Tuple of all the matrix indices that will be used to run simulations."""

        if not self.element_mappings:
            return self.matrix_indices_source

        # all the (i, j) pairs in `S_ij` that are tagged as covered by `element_mappings`
        elements_determined_by_map = [element_out for (_, element_out, _) in self.element_mappings]

        # loop through rows of the full s matrix and record rows that still need running.
        source_indices_needed = []
        for col_index in self.matrix_indices_source:
            # loop through columns and keep track of whether each element is covered by mapping.
            matrix_elements_covered = []
            for row_index in self.matrix_indices_monitor:
                element = (row_index, col_index)
                element_covered_by_map = element in elements_determined_by_map
                matrix_elements_covered.append(element_covered_by_map)

            # if any matrix elements in row still not covered by map, a source is needed for row.
            if not all(matrix_elements_covered):
                source_indices_needed.append(col_index)

        return source_indices_needed

    def _shift_value_signed(self, port: Union[Port, WavePort]) -> float:
        """How far (signed) to shift the source from the monitor."""

        return _shift_value_signed(
            obj=port,
            grid=self.simulation.grid,
            bounds=self.simulation.bounds,
            direction=port.direction,
            shift=-2,
            name=f"Port {port.name}",
        )

    unique_port_names = assert_unique_names("ports")

    def run(
        self,
        path_dir: str = DEFAULT_DATA_DIR,
        *,
        folder_name: str = "default",
        callback_url: Optional[str] = None,
        verbose: bool = True,
        solver_version: Optional[str] = None,
        pay_type: Union[PayType, str] = "AUTO",
        priority: Optional[int] = None,
        local_gradient: bool = False,
        max_num_adjoint_per_fwd: int = MAX_NUM_ADJOINT_PER_FWD,
    ):
        log.warning(
            "'ComponentModeler.run()' is deprecated and will be removed in a future release. "
            "Use web.run(modeler) instead. 'web.run' returns a 'ComponentModelerData' object; "
            "get the scattering matrix via 'data.smatrix()'.",
            log_once=True,
        )
        from tidy3d.plugins.smatrix.run import _run_local

        data = _run_local(
            self,
            path_dir=path_dir,
            folder_name=folder_name,
            callback_url=callback_url,
            verbose=verbose,
            solver_version=solver_version,
            pay_type=pay_type,
            priority=priority,
            local_gradient=local_gradient,
            max_num_adjoint_per_fwd=max_num_adjoint_per_fwd,
        )
        return data.smatrix()


AbstractComponentModeler.update_forward_refs()



from __future__ import annotations

from collections.abc import Iterable
from typing import Callable, Literal, Union

import autograd.numpy as np
import numpy as onp
from autograd import jacobian
from autograd.extend import defvjp, primitive
from autograd.scipy.signal import convolve as convolve_ag
from autograd.scipy.special import logsumexp
from autograd.tracer import getval
from numpy.lib.stride_tricks import sliding_window_view
from numpy.typing import NDArray

from tidy3d.components.autograd.functions import add_at, interpn, trapz

from .types import PaddingType

__all__ = [
    "add_at",
    "convolve",
    "grey_closing",
    "grey_dilation",
    "grey_erosion",
    "grey_opening",
    "interpn",
    "morphological_gradient",
    "morphological_gradient_external",
    "morphological_gradient_internal",
    "pad",
    "rescale",
    "smooth_max",
    "smooth_min",
    "threshold",
    "trapz",
]


def _get_pad_indices(
    n: int,
    pad_width: tuple[int, int],
    *,
    mode: PaddingType,
    numpy_module,
) -> NDArray:
    """Compute the indices to pad an array along a single axis based on the padding mode.

    Parameters
    ----------
    n : int
        The size of the axis to pad.
    pad_width : Tuple[int, int]
        The number of values padded to the edges of the axis.
    mode : PaddingType
        The padding mode to use.
    numpy_module : module
        The numpy module to use (either `numpy` or `autograd.numpy`).

    Returns
    -------
    np.ndarray
        The indices for padding along the axis.
    """
    total_pad = sum(pad_width)
    if n == 0:
        return numpy_module.zeros(total_pad, dtype=int)

    idx = numpy_module.arange(-pad_width[0], n + pad_width[1])

    if mode == "constant":
        return idx
    if mode == "edge":
        return numpy_module.clip(idx, 0, n - 1)
    if mode == "reflect":
        period = 2 * n - 2 if n > 1 else 1
        idx = numpy_module.mod(idx, period)
        return numpy_module.where(idx >= n, period - idx, idx)
    if mode == "symmetric":
        period = 2 * n if n > 1 else 1
        idx = numpy_module.mod(idx, period)
        return numpy_module.where(idx >= n, period - idx - 1, idx)
    if mode == "wrap":
        return numpy_module.mod(idx, n)

    raise ValueError(f"Unsupported padding mode: {mode}")


def pad(
    array: NDArray,
    pad_width: Union[int, tuple[int, int]],
    *,
    mode: PaddingType = "constant",
    axis: Union[int, Iterable[int], None] = None,
    constant_value: float = 0.0,
) -> NDArray:
    """Pad an array along specified axes with a given mode and padding width.

    Parameters
    ----------
    array : np.ndarray
        The input array to pad.
    pad_width : Union[int, Tuple[int, int]]
        The number of values padded to the edges of each axis. If an integer is provided,
        it is used for both the left and right sides. If a tuple is provided, it specifies
        the padding for the left and right sides respectively.
    mode : PaddingType = "constant"
        The padding mode to use.
    axis : Union[int, Iterable[int], None] = None
        The axis or axes along which to pad. If None, padding is applied to all axes.
    constant_value : float = 0.0
        The value to set the padded values for "constant" mode.

    Returns
    -------
    np.ndarray
        The padded array.

    Raises
    ------
    ValueError
        If the padding width has more than two elements or if padding is negative.
    IndexError
        If an axis is out of range for the array dimensions.
    """
    pad_width = np.atleast_1d(pad_width)
    if pad_width.size > 2:
        raise ValueError(f"Padding width must have one or two elements, got {pad_width.size}.")
    pad_tuple = (pad_width[0], pad_width[0]) if pad_width.size == 1 else tuple(pad_width)

    if any(p < 0 for p in pad_tuple):
        raise ValueError("Padding must be non-negative.")
    if all(p == 0 for p in pad_tuple):
        return array

    axes = range(array.ndim) if axis is None else [axis] if isinstance(axis, int) else axis
    axes = [ax + array.ndim if ax < 0 else ax for ax in axes]
    if any(ax < 0 or ax >= array.ndim for ax in axes):
        raise IndexError(f"Axis out of range for array with {array.ndim} dimensions.")

    result = array
    for ax in axes:
        if mode == "constant":
            padding = [(0, 0)] * result.ndim
            padding[ax] = pad_tuple
            result = np.pad(result, padding, mode="constant", constant_values=constant_value)
        else:
            idx = _get_pad_indices(result.shape[ax], pad_tuple, mode=mode, numpy_module=np)
            indexer = [slice(None)] * result.ndim
            indexer[ax] = idx
            result = result[tuple(indexer)]
    return result


def convolve(
    array: NDArray,
    kernel: NDArray,
    *,
    padding: PaddingType = "constant",
    axes: Union[tuple[list[int], list[int]], None] = None,
    mode: Literal["full", "valid", "same"] = "same",
) -> NDArray:
    """Convolve an array with a given kernel.

    Parameters
    ----------
    array : np.ndarray
        The input array to be convolved.
    kernel : np.ndarray
        The kernel to convolve with the input array. All dimensions of the kernel must be odd.
    padding : PaddingType = "constant"
        The padding mode to use.
    axes : Union[Tuple[List[int], List[int]], None] = None
        The axes along which to perform the convolution.
    mode : Literal["full", "valid", "same"] = "same"
        The convolution mode.

    Returns
    -------
    np.ndarray
        The result of the convolution.

    Raises
    ------
    ValueError
        If any dimension of the kernel is even.
        If the dimensions of the kernel do not match the dimensions of the array.
    """
    if any(k % 2 == 0 for k in kernel.shape):
        raise ValueError(f"All kernel dimensions must be odd, got {kernel.shape}.")

    if kernel.ndim != array.ndim and axes is None:
        raise ValueError(
            f"Kernel dimensions must match array dimensions, got kernel {kernel.shape} and array {array.shape}."
        )

    if mode in ("same", "full"):
        kernel_dims = kernel.shape if axes is None else [kernel.shape[d] for d in axes[1]]
        pad_widths = [(ks // 2, ks // 2) for ks in kernel_dims]
        for axis, pad_width in enumerate(pad_widths):
            array = pad(array, pad_width, mode=padding, axis=axis)
        mode = "valid" if mode == "same" else mode

    return convolve_ag(array, kernel, axes=axes, mode=mode)


def _get_footprint(size, structure, maxval):
    """Helper to generate the morphological footprint from size or structure."""
    if size is None and structure is None:
        raise ValueError("Either size or structure must be provided.")
    if size is not None and structure is not None:
        raise ValueError("Cannot specify both size and structure.")
    if structure is None:
        size_np = onp.atleast_1d(size)
        shape = (size_np[0], size_np[-1]) if size_np.size > 1 else (size_np[0], size_np[0])
        nb = onp.zeros(shape)
    else:
        structure_np = getval(structure)
        nb = onp.copy(structure_np)
        nb[structure_np == 0] = -maxval
    if nb.shape[0] % 2 == 0 or nb.shape[1] % 2 == 0:
        raise ValueError(f"Structuring element dimensions must be odd, got {nb.shape}.")
    return nb


@primitive
def grey_dilation(
    array: NDArray,
    size: Union[int, tuple[int, int], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey dilation on an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to perform grey dilation on.
    size : Union[Union[int, tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
        If a single integer is provided, a square structuring element is created.
        For 1D arrays, use a tuple (size, 1) or (1, size) for horizontal or vertical operations.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
        For 1D operations on 2D arrays, use a 2D structure with one dimension being 1.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The result of the grey dilation operation.

    Raises
    ------
    ValueError
        If both `size` and `structure` are None, or if the structuring element has even dimensions.
    """
    nb = _get_footprint(size, structure, maxval)
    h, w = nb.shape

    padded_array = pad(array, (h // 2, h // 2), mode=mode, axis=0)
    padded_array = pad(padded_array, (w // 2, w // 2), mode=mode, axis=1)

    padded_array_np = getval(padded_array)

    windows = sliding_window_view(padded_array_np, window_shape=(h, w))
    dilated_windows = windows + nb
    return onp.max(dilated_windows, axis=(-2, -1))


def _vjp_maker_dilation(ans, array, size=None, structure=None, *, mode="reflect", maxval=1e4):
    """VJP for the custom grey_dilation primitive."""
    nb = _get_footprint(size, structure, maxval)
    h, w = nb.shape

    padded_array = pad(array, (h // 2, h // 2), mode=mode, axis=0)
    padded_array = pad(padded_array, (w // 2, w // 2), mode=mode, axis=1)

    padded_array_np = getval(padded_array)
    in_h, in_w = getval(array).shape

    windows = sliding_window_view(padded_array_np, window_shape=(h, w))
    dilated_windows = windows + nb

    output_reshaped = ans[..., None, None]
    is_max_mask = (dilated_windows == output_reshaped).astype(onp.float64)

    # normalize the gradient for cases where multiple elements are the maximum.
    # When multiple elements in a window equal the maximum value, the gradient
    # is distributed equally among them. This ensures gradient conservation.
    # Note: Values can never exceed maxval in the output since we add structure
    # values (capped at maxval) to the input array values.
    multiplicity = onp.sum(is_max_mask, axis=(-2, -1), keepdims=True)
    is_max_mask /= onp.maximum(multiplicity, 1)

    def vjp(g):
        g_reshaped = g[..., None, None]
        grad_windows = g_reshaped * is_max_mask

        grad_padded = onp.zeros_like(padded_array_np)

        # create broadcastable indices for the scatter-add operation
        i = onp.arange(in_h)[:, None, None, None]
        j = onp.arange(in_w)[None, :, None, None]
        u = onp.arange(h)[None, None, :, None]
        v = onp.arange(w)[None, None, None, :]

        onp.add.at(grad_padded, (i + u, j + v), grad_windows)

        pad_h, pad_w = h // 2, w // 2

        # for constant padding, we can just slice the gradient
        if mode == "constant":
            return grad_padded[pad_h : pad_h + in_h, pad_w : pad_w + in_w]

        # for other modes, we need to sum gradients from padded regions by unpadding each axis
        grad_unpadded_w = onp.zeros((in_h + 2 * pad_h, in_w))
        padded_indices_w = _get_pad_indices(in_w, (pad_w, pad_w), mode=mode, numpy_module=onp)
        row_indices_w = onp.arange(in_h + 2 * pad_h)[:, None]
        onp.add.at(grad_unpadded_w, (row_indices_w, padded_indices_w), grad_padded)

        grad_unpadded_hw = onp.zeros((in_h, in_w))
        padded_indices_h = _get_pad_indices(in_h, (pad_h, pad_h), mode=mode, numpy_module=onp)[
            :, None
        ]
        col_indices_h = onp.arange(in_w)[None, :]
        onp.add.at(grad_unpadded_hw, (padded_indices_h, col_indices_h), grad_unpadded_w)

        return grad_unpadded_hw

    return vjp


defvjp(grey_dilation, _vjp_maker_dilation, argnums=[0])


def grey_erosion(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey erosion on an array.

    This function is implemented via duality, calling `grey_dilation` internally.

    Parameters
    ----------
    array : np.ndarray
        The input array to perform grey erosion on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The result of the grey erosion operation.
    """
    if structure is not None:
        structure = structure[::-1, ::-1]

    return -grey_dilation(
        -array,
        size=size,
        structure=structure,
        mode=mode,
        maxval=maxval,
    )


def grey_opening(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey opening on an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to perform grey opening on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The result of the grey opening operation.
    """
    array = grey_erosion(array, size, structure, mode=mode, maxval=maxval)
    array = grey_dilation(array, size, structure, mode=mode, maxval=maxval)
    return array


def grey_closing(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Perform grey closing on an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to perform grey closing on.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The result of the grey closing operation.
    """
    array = grey_dilation(array, size, structure, mode=mode, maxval=maxval)
    array = grey_erosion(array, size, structure, mode=mode, maxval=maxval)
    return array


def morphological_gradient(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the morphological gradient of an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to compute the morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The morphological gradient of the input array.
    """
    return grey_dilation(array, size, structure, mode=mode, maxval=maxval) - grey_erosion(
        array, size, structure, mode=mode, maxval=maxval
    )


def morphological_gradient_internal(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the internal morphological gradient of an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to compute the internal morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The internal morphological gradient of the input array.
    """
    return array - grey_erosion(array, size, structure, mode=mode, maxval=maxval)


def morphological_gradient_external(
    array: NDArray,
    size: Union[Union[int, tuple[int, int]], None] = None,
    structure: Union[NDArray, None] = None,
    *,
    mode: PaddingType = "reflect",
    maxval: float = 1e4,
) -> NDArray:
    """Compute the external morphological gradient of an array.

    Parameters
    ----------
    array : np.ndarray
        The input array to compute the external morphological gradient of.
    size : Union[Union[int, Tuple[int, int]], None] = None
        The size of the structuring element. If None, `structure` must be provided.
    structure : Union[np.ndarray, None] = None
        The structuring element. If None, `size` must be provided.
    mode : PaddingType = "reflect"
        The padding mode to use.
    maxval : float = 1e4
        Value to assume for infinite elements in the kernel.

    Returns
    -------
    np.ndarray
        The external morphological gradient of the input array.
    """
    return grey_dilation(array, size, structure, mode=mode, maxval=maxval) - array


def rescale(
    array: NDArray, out_min: float, out_max: float, in_min: float = 0.0, in_max: float = 1.0
) -> NDArray:
    """
    Rescale an array from an arbitrary input range to an arbitrary output range.

    Parameters
    ----------
    array : np.ndarray
        The input array to be rescaled.
    out_min : float
        The minimum value of the output range.
    out_max : float
        The maximum value of the output range.
    in_min : float = 0.0
        The minimum value of the input range.
    in_max : float = 1.0
        The maximum value of the input range.

    Returns
    -------
    np.ndarray
        The rescaled array.
    """

    if in_min == in_max:
        raise ValueError(
            f"'in_min' ({in_min}) must not be equal to 'in_max' ({in_max}) "
            "to avoid division by zero."
        )
    if out_min >= out_max:
        raise ValueError(f"'out_min' ({out_min}) must be less than 'out_max' ({out_max}).")
    if in_min >= in_max:
        raise ValueError(f"'in_min' ({in_min}) must be less than 'in_max' ({in_max}).")

    scaled = (array - in_min) / (in_max - in_min)
    return scaled * (out_max - out_min) + out_min


def threshold(
    array: NDArray, vmin: float = 0.0, vmax: float = 1.0, level: Union[float, None] = None
) -> NDArray:
    """Apply a threshold to an array, setting values below the threshold to `vmin` and values above to `vmax`.

    Parameters
    ----------
    array : np.ndarray
        The input array to be thresholded.
    vmin : float = 0.0
        The value to assign to elements below the threshold.
    vmax : float = 1.0
        The value to assign to elements above the threshold.
    level : Union[float, None] = None
        The threshold level. If None, the threshold is set to the midpoint between `vmin` and `vmax`.

    Returns
    -------
    np.ndarray
        The thresholded array.
    """
    if vmin >= vmax:
        raise ValueError(
            f"Invalid threshold range: 'vmin' ({vmin}) must be smaller than 'vmax' ({vmax})."
        )

    if level is None:
        level = (vmin + vmax) / 2
    elif not (vmin <= level <= vmax):
        raise ValueError(
            f"Invalid threshold level: 'level' ({level}) must be "
            f"between 'vmin' ({vmin}) and 'vmax' ({vmax})."
        )

    return np.where(array < level, vmin, vmax)


def smooth_max(
    x: NDArray, tau: float = 1.0, axis: Union[int, tuple[int, ...], None] = None
) -> float:
    """Compute the smooth maximum of an array using temperature parameter tau.

    Parameters
    ----------
    x : np.ndarray
        Input array.
    tau : float = 1.0
        Temperature parameter controlling smoothness. Larger values make the maximum smoother.
    axis : Union[int, Tuple[int, ...], None] = None
        Axis or axes over which the smooth maximum is computed. By default, the smooth maximum is computed over the entire array.

    Returns
    -------
    np.ndarray
        The smooth maximum of the input array.
    """
    return tau * logsumexp(x / tau, axis=axis)


def smooth_min(
    x: NDArray, tau: float = 1.0, axis: Union[int, tuple[int, ...], None] = None
) -> float:
    """Compute the smooth minimum of an array using temperature parameter tau.

    Parameters
    ----------
    x : np.ndarray
        Input array.
    tau : float = 1.0
        Temperature parameter controlling smoothness. Larger values make the minimum smoother.
    axis : Union[int, Tuple[int, ...], None] = None
        Axis or axes over which the smooth minimum is computed. By default, the smooth minimum is computed over the entire array.

    Returns
    -------
    np.ndarray
        The smooth minimum of the input array.
    """
    return -smooth_max(-x, tau, axis=axis)


def least_squares(
    func: Callable[[NDArray, float], NDArray],
    x: NDArray,
    y: NDArray,
    initial_guess: tuple[float, ...],
    max_iterations: int = 100,
    tol: float = 1e-6,
) -> NDArray:
    """Perform least squares fitting to find the best-fit parameters for a model function.

    Parameters
    ----------
    func : Callable[[np.ndarray, float], np.ndarray]
        The model function to fit. It should accept the independent variable `x` and a tuple of parameters,
        and return the predicted dependent variable values.
    x : np.ndarray
        Independent variable data.
    y : np.ndarray
        Dependent variable data.
    initial_guess : Tuple[float, ...]
        Initial guess for the parameters to be optimized.
    max_iterations : int = 100
        Maximum number of iterations for the optimization process.
    tol : float = 1e-6
        Tolerance for convergence. The optimization stops when the change in parameters is below this threshold.

    Returns
    -------
    np.ndarray
        The optimized parameters that best fit the model to the data.

    Raises
    ------
    np.linalg.LinAlgError
        If the optimization does not converge within the specified number of iterations.

    Example
    -------
    >>> import numpy as np
    >>> def linear_model(x, a, b):
    ...     return a * x + b
    >>> x_data = np.linspace(0, 10, 50)
    >>> y_data = 2.0 * x_data - 3.0
    >>> initial_guess = (0.0, 0.0)
    >>> params = least_squares(linear_model, x_data, y_data, initial_guess)
    >>> print(params)
    [ 2. -3.]
    """
    params = np.array(initial_guess, dtype="f8")
    jac = jacobian(lambda params: func(x, *params))

    for _ in range(max_iterations):
        residuals = y - func(x, *params)
        jacvec = jac(params)
        pseudo_inv = np.linalg.pinv(jacvec)
        delta = np.dot(pseudo_inv, residuals)
        params = params + delta
        if np.linalg.norm(delta) < tol:
            break

    return params



"""Defines an abstract base for electromagnetic sources."""

from __future__ import annotations

from abc import ABC
from typing import Optional

import pydantic.v1 as pydantic

from tidy3d.components.base import cached_property
from tidy3d.components.base_sim.source import AbstractSource
from tidy3d.components.geometry.base import Box
from tidy3d.components.types import TYPE_TAG_STR, Ax
from tidy3d.components.validators import _assert_min_freq, _warn_unsupported_traced_argument
from tidy3d.components.viz import (
    ARROW_ALPHA,
    ARROW_COLOR_POLARIZATION,
    ARROW_COLOR_SOURCE,
    PlotParams,
    plot_params_source,
)

from .time import SourceTimeType


class Source(Box, AbstractSource, ABC):
    """Abstract base class for all sources."""

    source_time: SourceTimeType = pydantic.Field(
        ...,
        title="Source Time",
        description="Specification of the source time-dependence.",
        discriminator=TYPE_TAG_STR,
    )

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a Source object."""
        return plot_params_source

    @cached_property
    def geometry(self) -> Box:
        """:class:`Box` representation of source."""

        return Box(center=self.center, size=self.size)

    @cached_property
    def _injection_axis(self):
        """Injection axis of the source."""
        return

    @cached_property
    def _dir_vector(self) -> tuple[float, float, float]:
        """Returns a vector indicating the source direction for arrow plotting, if not None."""
        return None

    @cached_property
    def _pol_vector(self) -> tuple[float, float, float]:
        """Returns a vector indicating the source polarization for arrow plotting, if not None."""
        return None

    _warn_traced_center = _warn_unsupported_traced_argument("center")
    _warn_traced_size = _warn_unsupported_traced_argument("size")

    @pydantic.validator("source_time", always=True)
    def _freqs_lower_bound(cls, val):
        """Raise validation error if central frequency is too low."""
        _assert_min_freq(val.freq0, msg_start="'source_time.freq0'")
        return val

    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **patch_kwargs,
    ) -> Ax:
        """Plot this source."""

        kwargs_arrow_base = patch_kwargs.pop("arrow_base", None)

        # call the `Source.plot()` function first.
        ax = Box.plot(self, x=x, y=y, z=z, ax=ax, **patch_kwargs)

        kwargs_alpha = patch_kwargs.get("alpha")
        arrow_alpha = ARROW_ALPHA if kwargs_alpha is None else kwargs_alpha

        # then add the arrow based on the propagation direction
        if self._dir_vector is not None:
            bend_radius = None
            bend_axis = None
            if hasattr(self, "mode_spec") and self.mode_spec.bend_radius is not None:
                bend_radius = self.mode_spec.bend_radius
                bend_axis = self._bend_axis
                sign = 1 if self.direction == "+" else -1
                # Curvature has to be reversed because of ploting coordinates
                if (self.size.index(0), bend_axis) in [(1, 2), (2, 0), (2, 1)]:
                    bend_radius *= -sign
                else:
                    bend_radius *= sign

            ax = self._plot_arrow(
                x=x,
                y=y,
                z=z,
                ax=ax,
                direction=self._dir_vector,
                bend_radius=bend_radius,
                bend_axis=bend_axis,
                color=ARROW_COLOR_SOURCE,
                alpha=arrow_alpha,
                both_dirs=False,
                arrow_base=kwargs_arrow_base,
            )

        if self._pol_vector is not None:
            ax = self._plot_arrow(
                x=x,
                y=y,
                z=z,
                ax=ax,
                direction=self._pol_vector,
                color=ARROW_COLOR_POLARIZATION,
                alpha=arrow_alpha,
                both_dirs=False,
                arrow_base=kwargs_arrow_base,
            )

        return ax



"""Defines heat material specifications"""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.components.spice.sources.types import CurrentSourceType, VoltageSourceType
from tidy3d.components.tcad.boundary.abstract import HeatChargeBC
from tidy3d.constants import CURRENT_DENSITY, VOLT


class VoltageBC(HeatChargeBC):
    """
    Constant electric potential (voltage) :math:`= \\text{V}` boundary condition.
    Sets a potential at the specified boundary.

    Notes
    -----

        In charge simulations it also accepts an array of voltages.
        In this case, a solution for each of these voltages will
        be computed.

    Example
    -------
    >>> import tidy3d as td
    >>> voltage_source = td.DCVoltageSource(voltage=1)
    >>> voltage_bc = td.VoltageBC(source=voltage_source)
    """

    source: VoltageSourceType = pd.Field(
        ...,
        title="Voltage",
        description="Electric potential to be applied at the specified boundary.",
        units=VOLT,
    )


class CurrentBC(HeatChargeBC):
    """
    Current boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> current_source = td.DCCurrentSource(current=1)
    >>> current_bc = CurrentBC(source=current_source)
    """

    source: CurrentSourceType = pd.Field(
        ...,
        title="Current Source",
        description="A current source",
        units=CURRENT_DENSITY,
    )
    # TODO translation between currentsource amps and currentdensity, why not amps here?


class InsulatingBC(HeatChargeBC):
    """Insulation boundary condition.

    Notes
    -----

        Ensures the electric potential to the normal :math:`\\nabla \\psi \\cdot \\mathbf{n}  = 0` as well as the
        surface recombination current density :math:`J_s = \\mathbf{J} \\cdot \\mathbf{n} = 0` are set to zero where
        the current density is :math:`\\mathbf{J_n}` and the normal vector is :math:`\\mathbf{n}`

    Example
    -------
    >>> bc = InsulatingBC()
    """



"""Data structures for post-processing terminal component simulations to calculate S-matrices."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.data.data_array import FreqDataArray
from tidy3d.components.data.monitor_data import MonitorData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.microwave.data.monitor_data import AntennaMetricsData
from tidy3d.log import log
from tidy3d.plugins.smatrix.component_modelers.terminal import TerminalComponentModeler
from tidy3d.plugins.smatrix.data.base import AbstractComponentModelerData
from tidy3d.plugins.smatrix.data.data_array import PortDataArray, TerminalPortDataArray
from tidy3d.plugins.smatrix.ports.types import TerminalPortType
from tidy3d.plugins.smatrix.types import SParamDef
from tidy3d.plugins.smatrix.utils import (
    ab_to_s,
    check_port_impedance_sign,
    compute_F,
    compute_port_VI,
    compute_power_delivered_by_port,
    compute_power_wave_amplitudes,
    s_to_z,
)


class MicrowaveSMatrixData(Tidy3dBaseModel):
    """Stores the computed S-matrix and reference impedances for the terminal ports."""

    port_reference_impedances: Optional[PortDataArray] = pd.Field(
        None,
        title="Port Reference Impedances",
        description="Reference impedance for each port used in the S-parameter calculation. This is optional and may not be present if not specified or computed.",
    )

    data: TerminalPortDataArray = pd.Field(
        ...,
        title="S-Matrix Data",
        description="An array containing the computed S-matrix of the device. The data is organized by terminal ports, representing the scattering parameters between them.",
    )

    s_param_def: SParamDef = pd.Field(
        "pseudo",
        title="Scattering Parameter Definition",
        description="Whether scattering parameters are defined using the 'pseudo' or 'power' wave definitions.",
    )


class TerminalComponentModelerData(AbstractComponentModelerData):
    """
    Data associated with a :class:`.TerminalComponentModeler` simulation run.


    Notes
    -----

    This class serves as a data container for the results of a component modeler simulation,
    with the original simulation definition, and port simulation data, and the solver log.


    **References**

    .. [1]  R. B. Marks and D. F. Williams, "A general waveguide circuit theory,"
            J. Res. Natl. Inst. Stand. Technol., vol. 97, pp. 533, 1992.

    .. [2]  D. M. Pozar, Microwave Engineering, 4th ed. Hoboken, NJ, USA:
            John Wiley & Sons, 2012.
    """

    modeler: TerminalComponentModeler = pd.Field(
        ...,
        title="TerminalComponentModeler",
        description="The original :class:`.TerminalComponentModeler` object that defines the simulation setup "
        "and from which this data was generated.",
    )

    def smatrix(
        self,
        assume_ideal_excitation: Optional[bool] = None,
        s_param_def: Optional[SParamDef] = None,
    ) -> MicrowaveSMatrixData:
        """Computes and returns the S-matrix and port reference impedances.

        Parameters
        ----------
        assume_ideal_excitation: If ``True``, assumes that exciting one port
            does not produce incident waves at other ports. This simplifies the
            S-matrix calculation and is required if not all ports are excited. If not
            provided, ``modeler.assume_ideal_excitation`` is used.
        s_param_def: The definition of S-parameters to use, determining whether
            "pseudo waves" or "power waves" are calculated. If not provided,
            ``modeler.s_param_def`` is used.

        Returns
        -------
        :class:`.MicrowaveSMatrixData`
            Container with the computed S-parameters and the port reference impedances.
        """
        from tidy3d.plugins.smatrix.analysis.terminal import terminal_construct_smatrix

        terminal_port_data = terminal_construct_smatrix(
            modeler_data=self,
            assume_ideal_excitation=assume_ideal_excitation
            if (assume_ideal_excitation is not None)
            else self.modeler.assume_ideal_excitation,
            s_param_def=s_param_def if (s_param_def is not None) else self.modeler.s_param_def,
        )
        smatrix_data = MicrowaveSMatrixData(
            data=terminal_port_data,
            port_reference_impedances=self.port_reference_impedances,
            s_param_def=s_param_def if (s_param_def is not None) else self.modeler.s_param_def,
        )
        return smatrix_data

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values

    def _monitor_data_at_port_amplitude(
        self,
        port: TerminalPortType,
        monitor_name: str,
        a_port: Union[FreqDataArray, complex],
    ) -> MonitorData:
        """Normalize the monitor data to a desired complex amplitude of a port,
        represented by ``a_port``, where :math:`\frac{1}{2}|a|^2` is the power
        incident from the port into the system.
        """
        sim_data_port = self.data[self.modeler.get_task_name(port)]
        monitor_data = sim_data_port[monitor_name]
        a_raw, _ = self.compute_power_wave_amplitudes_at_each_port(sim_data=sim_data_port)
        a_raw_port = a_raw.sel(port=self.modeler.network_index(port))
        if not isinstance(a_port, FreqDataArray):
            freqs = list(monitor_data.monitor.freqs)
            array_vals = a_port * np.ones(len(freqs))
            a_port = FreqDataArray(array_vals, coords={"f": freqs})
        scale_array = a_port / a_raw_port
        return monitor_data.scale_fields_by_freq_array(scale_array, method="nearest")

    def get_antenna_metrics_data(
        self,
        port_amplitudes: Optional[dict[str, complex]] = None,
        monitor_name: Optional[str] = None,
    ) -> AntennaMetricsData:
        """Calculate antenna parameters using superposition of fields from multiple port excitations.

        The method computes the radiated far fields and port excitation power wave amplitudes
        for a superposition of port excitations, which can be used to analyze antenna radiation
        characteristics.

        Parameters
        ----------
        port_amplitudes : dict[str, complex]
            Dictionary mapping port names to their desired excitation amplitudes. For each port,
            :math:`\\frac{1}{2}|a|^2` represents the incident power from that port into the system.
            If None, uses only the first port without any scaling of the raw simulation data.  When ``None``
            is passed as a port amplitude, the raw simulation data is used for that port. Note that in this method ``a`` represents
            the incident wave amplitude using the power wave definition in [2].
        monitor_name : str
            Name of the :class:`.DirectivityMonitor` to use for calculating far fields.
            If None, uses the first monitor in `radiation_monitors`.

        Returns
        -------
        :class:`.AntennaMetricsData`
            Container with antenna parameters including directivity, gain, and radiation efficiency,
            computed from the superposition of fields from all excited ports.
        """
        from tidy3d.plugins.smatrix.analysis.antenna import get_antenna_metrics_data

        antenna_metrics_data = get_antenna_metrics_data(
            terminal_component_modeler_data=self,
            port_amplitudes=port_amplitudes,
            monitor_name=monitor_name,
        )
        return antenna_metrics_data

    @cached_property
    def port_reference_impedances(self) -> PortDataArray:
        """Calculates the reference impedance for each port across all frequencies.

        This function determines the characteristic impedance for every port defined
        in the modeler. It handles two types of ports differently: for a
        :class:`.WavePort`, the impedance is frequency-dependent and computed from
        modal properties, while for other types like :class:`.LumpedPort`, the
        impedance is a user-defined constant value.

        Returns:
            A data array containing the complex impedance for each port at each
            frequency.
        """
        from tidy3d.plugins.smatrix.analysis.terminal import port_reference_impedances

        return port_reference_impedances(self)

    def compute_wave_amplitudes_at_each_port(
        self,
        sim_data: SimulationData,
        port_reference_impedances: Optional[PortDataArray] = None,
        s_param_def: SParamDef = "pseudo",
    ) -> tuple[PortDataArray, PortDataArray]:
        """Compute the incident and reflected amplitudes at each port.
        The computed amplitudes have not been normalized.

        Parameters
        ----------
        sim_data : :class:`.SimulationData`
            Results from the simulation.
        port_reference_impedances : :class:`.PortDataArray`, optional
            Reference impedance at each port. If not provided, it is computed from the cached
            property :meth:`.port_reference_impedances`. Defaults to ``None``.
        s_param_def : SParamDef
            The type of waves computed, either pseudo waves defined by Equation 53 and Equation 54 in [1],
            or power waves defined by Equation 4.67 in [2].

        Returns
        -------
        tuple[:class:`.PortDataArray`, :class:`.PortDataArray`]
            Incident (a) and reflected (b) wave amplitudes at each port.
        """
        from tidy3d.plugins.smatrix.analysis.terminal import compute_wave_amplitudes_at_each_port

        port_reference_impedances_i = (
            port_reference_impedances
            if port_reference_impedances is not None
            else self.port_reference_impedances
        )

        return compute_wave_amplitudes_at_each_port(
            modeler=self.modeler,
            port_reference_impedances=port_reference_impedances_i,
            sim_data=sim_data,
            s_param_def=s_param_def,
        )

    def compute_power_wave_amplitudes_at_each_port(
        self,
        sim_data: SimulationData,
        port_reference_impedances: Optional[PortDataArray] = None,
    ) -> tuple[PortDataArray, PortDataArray]:
        """Compute the incident and reflected power wave amplitudes at each port.
        The computed amplitudes have not been normalized.

        Parameters
        ----------
        sim_data : :class:`.SimulationData`
            Results from the simulation.
        port_reference_impedances : :class:`.PortDataArray`, optional
            Reference impedance at each port. If not provided, it is computed from the cached
            property :meth:`.port_reference_impedances`. Defaults to ``None``.

        Returns
        -------
        tuple[:class:`.PortDataArray`, :class:`.PortDataArray`]
            Incident (a) and reflected (b) power wave amplitudes at each port.
        """
        from tidy3d.plugins.smatrix.analysis.terminal import (
            compute_power_wave_amplitudes_at_each_port,
        )

        port_reference_impedances_i = (
            port_reference_impedances
            if port_reference_impedances is not None
            else self.port_reference_impedances
        )

        return compute_power_wave_amplitudes_at_each_port(
            modeler=self.modeler,
            port_reference_impedances=port_reference_impedances_i,
            sim_data=sim_data,
        )

    def s_to_z(
        self,
        reference: Union[complex, PortDataArray],
        assume_ideal_excitation: Optional[bool] = None,
        s_param_def: SParamDef = "pseudo",
    ) -> TerminalPortDataArray:
        """Converts the S-matrix to the Z-matrix using a specified reference impedance.

        This method first computes the S-matrix of the device and then transforms it into the
        corresponding impedance matrix (Z-matrix). The conversion can be performed using either a
        single, uniform reference impedance for all ports or a more general set of per-port,
        frequency-dependent reference impedances.

        This method :meth:`.TerminalComponentModelerData.s_to_z` is called on a
        :class:`.TerminalComponentModelerData` object, which contains the S-matrix and other
        simulation data internally.

        Parameters
        ----------
        reference : Union[complex, :class:`.PortDataArray`]
            The reference impedance(s) to use for the conversion. If a single complex value is
            provided, it is assumed to be the reference impedance for all ports. If a
            :class:`.PortDataArray` is given, it should contain the specific reference
            impedance for each port.
        assume_ideal_excitation: If ``True``, assumes that exciting one port
            does not produce incident waves at other ports. This simplifies the
            S-matrix calculation and is required if not all ports are excited. If not
            provided, ``modeler.assume_ideal_excitation`` is used.
        s_param_def : SParamDef, optional
            The definition of the scattering parameters used in the S-matrix calculation.
            This can be either "pseudo" for pseudo waves (see [1]) or "power" for power
            waves (see [2]). Defaults to "pseudo".

        Returns
        -------
        DataArray
            The computed impedance (Z) matrix, with dimensions corresponding to the ports of
            the device.

        Examples
        --------
        >>> z_matrix = component_modeler_data.s_to_z(reference=50) # doctest: +SKIP
        >>> z_11 = z_matrix.sel(port_out="port_1@0", port_in="port_1@0") # doctest: +SKIP

        See Also
        --------
        smatrix : Computes the scattering matrix.
        """
        s_matrix = self.smatrix(
            assume_ideal_excitation=assume_ideal_excitation, s_param_def=s_param_def
        )
        return s_to_z(s_matrix=s_matrix.data, reference=reference, s_param_def=s_param_def)

    # Mirror Utils
    # So they can be reused elsewhere without a class reimport
    ab_to_s = staticmethod(ab_to_s)
    compute_F = staticmethod(compute_F)
    check_port_impedance_sign = staticmethod(check_port_impedance_sign)
    compute_port_VI = staticmethod(compute_port_VI)
    compute_power_wave_amplitudes = staticmethod(compute_power_wave_amplitudes)
    compute_power_delivered_by_port = staticmethod(compute_power_delivered_by_port)



"""Monitor level data, store the DataArrays associated with a single heat-charge monitor."""

from __future__ import annotations

from typing import Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.data.data_array import (
    DataArray,
    IndexedFieldVoltageDataArray,
    IndexedVoltageDataArray,
    PointDataArray,
    SpatialDataArray,
    SteadyVoltageDataArray,
)
from tidy3d.components.data.utils import TetrahedralGridDataset, TriangularGridDataset
from tidy3d.components.tcad.data.monitor_data.abstract import HeatChargeMonitorData
from tidy3d.components.tcad.monitors.charge import (
    SteadyCapacitanceMonitor,
    SteadyCurrentDensityMonitor,
    SteadyElectricFieldMonitor,
    SteadyEnergyBandMonitor,
    SteadyFreeCarrierMonitor,
    SteadyPotentialMonitor,
)
from tidy3d.components.types import TYPE_TAG_STR, Ax, annotate_type
from tidy3d.components.viz import add_ax_if_none
from tidy3d.exceptions import DataError

FieldDataset = Union[
    SpatialDataArray, annotate_type(Union[TriangularGridDataset, TetrahedralGridDataset])
]

UnstructuredFieldType = Union[TriangularGridDataset, TetrahedralGridDataset]


class SteadyPotentialData(HeatChargeMonitorData):
    """Stores electric potential :math:`\\psi` from a charge simulation."""

    monitor: SteadyPotentialMonitor = pd.Field(
        ...,
        title="Electric potential monitor",
        description="Electric potential monitor associated with a `charge` simulation.",
    )

    potential: FieldDataset = pd.Field(
        None,
        title="Electric potential series",
        description="Contains the electric potential series.",
    )

    @property
    def field_components(self) -> dict[str, DataArray]:
        """Maps the field components to their associated data."""
        return {"potential": self.potential}


class SteadyFreeCarrierData(HeatChargeMonitorData):
    """
    Stores free-carrier concentration in charge simulations.

    Notes
    -----

        This data contains the carrier concentrations: the amount of electrons and holes per unit volume as defined in the
        ``monitor``.
    """

    monitor: SteadyFreeCarrierMonitor = pd.Field(
        ...,
        title="Free carrier monitor",
        description="Free carrier data associated with a Charge simulation.",
    )

    electrons: UnstructuredFieldType = pd.Field(
        None,
        title="Electrons series",
        description=r"Contains the computed electrons concentration $n$.",
        discriminator=TYPE_TAG_STR,
    )
    # n = electrons

    holes: UnstructuredFieldType = pd.Field(
        None,
        title="Holes series",
        description=r"Contains the computed holes concentration $p$.",
        discriminator=TYPE_TAG_STR,
    )
    # p = holes

    @property
    def field_components(self) -> dict[str, DataArray]:
        """Maps the field components to their associated data."""
        return {"electrons": self.electrons, "holes": self.holes}

    @pd.root_validator(skip_on_failure=True)
    def check_correct_data_type(cls, values):
        """Issue error if incorrect data type is used"""

        mnt = values.get("monitor")
        field_data = {field: values.get(field) for field in ["electrons", "holes"]}

        for field, data in field_data.items():
            if isinstance(data, TetrahedralGridDataset) or isinstance(data, TriangularGridDataset):
                if not isinstance(data.values, IndexedVoltageDataArray):
                    raise ValueError(
                        f"In the data associated with monitor {mnt}, the field {field} does not contain "
                        "data associated to any voltage value."
                    )

        return values


class SteadyEnergyBandData(HeatChargeMonitorData):
    """
    Stores energy bands in charge simulations.

    Notes
    -----

        This data contains the energy bands data:
        Ec -> Energy of the bottom of the conduction band, [eV]
        Ev -> Energy of the top of the valence band, [eV]
        Ei -> Intrinsic Fermi level, [eV]
        Efn -> Quasi-Fermi level for electrons, [eV]
        Efp -> Quasi-Fermi level for holes, [eV]
        as defined in the  ``monitor``.
    """

    monitor: SteadyEnergyBandMonitor = pd.Field(
        ...,
        title="Energy band monitor",
        description="Energy bands data associated with a Charge simulation.",
    )

    Ec: UnstructuredFieldType = pd.Field(
        None,
        title="Conduction band series",
        description=r"Contains the computed energy of the bottom of the conduction band $Ec$.",
        discriminator=TYPE_TAG_STR,
    )

    Ev: UnstructuredFieldType = pd.Field(
        None,
        title="Valence band series",
        description=r"Contains the computed energy of the top of the valence band $Ec$.",
        discriminator=TYPE_TAG_STR,
    )

    Ei: UnstructuredFieldType = pd.Field(
        None,
        title="Intrinsic Fermi level series",
        description=r"Contains the computed intrinsic Fermi level for the material $Ei$.",
        discriminator=TYPE_TAG_STR,
    )

    Efn: UnstructuredFieldType = pd.Field(
        None,
        title="Electron's quasi-Fermi level series",
        description=r"Contains the computed quasi-Fermi level for electrons $Efn$.",
        discriminator=TYPE_TAG_STR,
    )

    Efp: UnstructuredFieldType = pd.Field(
        None,
        title="Hole's quasi-Fermi level series",
        description=r"Contains the computed quasi-Fermi level for holes $Efp$.",
        discriminator=TYPE_TAG_STR,
    )

    @property
    def field_components(self) -> dict[str, DataArray]:
        """Maps the field components to their associated data."""
        return {"Ec": self.Ec, "Ev": self.Ev, "Ei": self.Ei, "Efn": self.Efn, "Efp": self.Efp}

    @pd.root_validator(skip_on_failure=True)
    def check_correct_data_type(cls, values):
        """Issue error if incorrect data type is used"""

        mnt = values.get("monitor")
        field_data = {field: values.get(field) for field in ["Ec", "Ev", "Ei", "Efn", "Efp"]}

        for field, data in field_data.items():
            if isinstance(data, TetrahedralGridDataset) or isinstance(data, TriangularGridDataset):
                if not isinstance(data.values, IndexedVoltageDataArray):
                    raise ValueError(
                        f"In the data associated with monitor {mnt}, the field {field} does not contain "
                        "data associated to any voltage value."
                    )

        return values

    @add_ax_if_none
    def plot(self, ax: Ax = None, **sel_kwargs) -> Ax:
        """Plot the 1D cross-section of the energy bandgap diagram.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.Axes = None
            matplotlib axes to plot on, if not specified, one is created.
        sel_kwargs : keyword arguments used to perform ``.sel()`` selection in the monitor data.
            These kwargs can select over the spatial dimensions (``x``, ``y``, or ``z``)
            and the bias voltage (``voltage``).
            For the plotting to work appropriately, the resulting data after selection must contain
            only one coordinate with len > 1.
            Furthermore, these should be spatial coordinates (``x``, ``y``, or ``z``).
        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        selection_data = {}

        if ("voltage" not in sel_kwargs) and (self.Ec.values.coords.sizes["voltage"] > 1):
            raise DataError(
                "'voltage' is not selected for the plot with multiple voltage data points."
            )

        selection_data = {coord: sel_kwargs[coord] for coord in "xyz" if coord in sel_kwargs.keys()}
        N_coords = len(selection_data.keys())

        if "voltage" in sel_kwargs:
            selection_data["voltage"] = sel_kwargs["voltage"]

        if isinstance(self.Ec, TetrahedralGridDataset):
            if N_coords != 2:
                raise DataError(
                    "2 spatial coordinate values have to be defined to plot the 1D cross-section figure for a 3D dataset."
                )

        elif isinstance(self.Ec, TriangularGridDataset):
            if N_coords != 1:
                raise DataError(
                    "1 spatial coordinate value has to be defined to plot the 1D cross-section figure for a 2D dataset."
                )

            for index, coord_name in enumerate(["x", "y", "z"]):
                if coord_name in selection_data:
                    axis = index
                    continue

            if axis == self.Ec.normal_axis:
                raise DataError(
                    f"Triangular grid (normal: {self.Ec.normal_axis}) cannot be sliced by a parallel plane."
                )

        Ec_data = self.Ec
        Ev_data = self.Ev
        Ei_data = self.Ei
        Efn_data = self.Efn
        Efp_data = self.Efp

        for coord_name, coord_val in selection_data.items():
            Ec_data = Ec_data.sel(**{coord_name: coord_val}, method="nearest")
            Ev_data = Ev_data.sel(**{coord_name: coord_val}, method="nearest")
            Ei_data = Ei_data.sel(**{coord_name: coord_val}, method="nearest")
            Efn_data = Efn_data.sel(**{coord_name: coord_val}, method="nearest")
            Efp_data = Efp_data.sel(**{coord_name: coord_val}, method="nearest")

        Ec_data.plot(ax=ax, label="Ec")
        Ev_data.plot(ax=ax, label="Ev")
        Ei_data.plot(ax=ax, label="Ei")
        Efn_data.plot(ax=ax, label="Efn")
        Efp_data.plot(ax=ax, label="Efp")
        ax.legend()

        return ax


class SteadyCapacitanceData(HeatChargeMonitorData):
    """
    Class that stores capacitance data from a Charge simulation.

    Notes
    -----
        The small signal-capacitance of electrons :math:`C_n` and holes  :math:`C_p`  is computed from the charge due to
         electrons :math:`Q_n` and holes :math:`Q_p` at an applied voltage :math:`V` at a voltage difference
        :math:`\\Delta V` between two simulations.

        .. math::

            C_{n,p} = \\frac{Q_{n,p}(V + \\Delta V) - Q_{n,p}(V)}{\\Delta V}


    This is only computed when a voltage source with more than two sources is included within the simulation and determines the :math:`\\Delta V`.
    """

    monitor: SteadyCapacitanceMonitor = pd.Field(
        ...,
        title="Capacitance monitor",
        description="Capacitance data associated with a Charge simulation.",
    )

    hole_capacitance: SteadyVoltageDataArray = pd.Field(
        None,
        title="Hole capacitance",
        description=r"Small signal capacitance ($\frac{dQ_p}{dV}$) associated to the monitor.",
    )
    # C_p = hole_capacitance

    electron_capacitance: SteadyVoltageDataArray = pd.Field(
        None,
        title="Electron capacitance",
        description=r"Small signal capacitance ($\frac{dQn}{dV}$) associated to the monitor.",
    )
    # C_n = electron_capacitance

    @property
    def field_components(self) -> dict[str, UnstructuredFieldType]:
        """Maps the field components to their associated data."""
        return {}

    @property
    def symmetry_expanded_copy(self) -> SteadyCapacitanceData:
        """Return copy of self with symmetry applied."""
        num_symmetries = np.sum(np.array([1 if d > 0 else 0 for d in self.symmetry]))
        scaling_factor = np.power(2, num_symmetries)

        if self.hole_capacitance is None:
            new_hole_capacitance = None
        else:
            new_values = self.hole_capacitance.values * scaling_factor
            new_hole_capacitance = SteadyVoltageDataArray(
                data=new_values, coords=self.hole_capacitance.coords
            )

        if self.electron_capacitance is None:
            new_electron_capacitance = None
        else:
            new_values = self.electron_capacitance.values * scaling_factor
            new_electron_capacitance = SteadyVoltageDataArray(
                data=new_values, coords=self.electron_capacitance.coords
            )

        return self.updated_copy(
            hole_capacitance=new_hole_capacitance,
            electron_capacitance=new_electron_capacitance,
            symmetry=(0, 0, 0),
        )


class SteadyElectricFieldData(HeatChargeMonitorData):
    """
    Stores electric field :math:`\\vec{E}` from a Charge/Conduction simulation.

    Notes
    -----
        The electric field is computed as the negative gradient of the electric potential :math:`\\vec{E} = -\\nabla \\psi`.
        It is given in units of :math:`V/\\mu m` (Volts per micrometer).
    """

    monitor: SteadyElectricFieldMonitor = pd.Field(
        ...,
        title="Electric field monitor",
        description="Electric field data associated with a Charge/Conduction simulation.",
    )

    E: UnstructuredFieldType = pd.Field(
        None,
        title="Electric field",
        description=r"Contains the computed electric field in :math:`V/\\mu m`.",
        discriminator=TYPE_TAG_STR,
    )

    @property
    def field_components(self) -> dict[str, UnstructuredFieldType]:
        """Maps the field components to their associated data."""
        return {"E": self.E}

    @pd.root_validator(skip_on_failure=True)
    def check_correct_data_type(cls, values):
        """Issue error if incorrect data type is used"""

        mnt = values.get("monitor")
        E = values.get("E")

        if isinstance(E, TetrahedralGridDataset) or isinstance(E, TriangularGridDataset):
            AcceptedTypes = (IndexedFieldVoltageDataArray, PointDataArray)
            if not isinstance(E.values, AcceptedTypes):
                raise ValueError(
                    f"In the data associated with monitor {mnt}, must contain a field. This can be "
                    "defined with IndexedFieldVoltageDataArray or PointDataArray."
                )

        return values


class SteadyCurrentDensityData(HeatChargeMonitorData):
    """
    Stores current density :math:`\\vec{J}` from a Charge/Conduction simulation. It is given in
    units of :math:`A/\\mu m^2`
    """

    monitor: SteadyCurrentDensityMonitor = pd.Field(
        ...,
        title="Current density monitor",
        description="Current density data associated with a Charge/Conduction simulation.",
    )

    J: UnstructuredFieldType = pd.Field(
        None,
        title="Current density",
        description=r"Contains the computed current density in :math:`A/\\mu m^2`.",
        discriminator=TYPE_TAG_STR,
    )

    @property
    def field_components(self) -> dict[str, UnstructuredFieldType]:
        """Maps the field components to their associated data."""
        return {"J": self.J}

    @pd.root_validator(skip_on_failure=True)
    def check_correct_data_type(cls, values):
        """Issue error if incorrect data type is used"""

        mnt = values.get("monitor")
        J = values.get("J")

        if isinstance(J, TetrahedralGridDataset) or isinstance(J, TriangularGridDataset):
            AcceptedTypes = (IndexedFieldVoltageDataArray, PointDataArray)
            if not isinstance(J.values, AcceptedTypes):
                raise ValueError(
                    f"In the data associated with monitor {mnt}, must contain a field. This can be "
                    "defined with IndexedFieldVoltageDataArray or PointDataArray."
                )

        return values



"""Special types and validators used by adjoint plugin."""

from __future__ import annotations

from typing import Any, Union

import numpy as np

from tidy3d.components.types import _add_schema

# special handling if we cant import the JVPTracer in the future (so it doesn't break tidy3d).
try:
    from jax.interpreters.ad import JVPTracer
except ImportError:
    import tidy3d as td

    td.log.warning(
        "Could not import 'jax.interpreters.ad.JVPTracer'. "
        "As a temporary fix, 'jax'-traced floats will use 'typing.Any' in their type annotation. "
        "If you encounter this warning, please file an issue on the Tidy3D front end repository "
        "as it indicates that the 'adjoint' plugin will need to be upgraded."
    )
    JVPTracer = Any

from jax.numpy import ndarray as JaxArrayType

""" Define schema for these jax and numpy types."""


class NumpyArrayType(np.ndarray):
    """Subclass of ``np.ndarray`` with a schema defined for pydantic."""

    @classmethod
    def __modify_schema__(cls, field_schema):
        """Sets the schema of np.ndarray object."""

        schema = {
            "title": "npdarray",
            "type": "numpy.ndarray",
        }
        field_schema.update(schema)


_add_schema(JaxArrayType, title="JaxArray", field_type_str="jax.numpy.ndarray")

# if the ImportError didnt occur, add the schema
if JVPTracer is not Any:
    _add_schema(JVPTracer, title="JVPTracer", field_type_str="jax.interpreters.ad.JVPTracer")

# define types usable as floats including the jax tracers
JaxArrayLike = Union[NumpyArrayType, JaxArrayType]
JaxFloat = Union[float, JaxArrayLike, JVPTracer, object]
# note: object is included here because sometimes jax passes just `object` (i think when untraced)



from __future__ import annotations

from typing import Optional

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.material.solver_types import (
    ChargeMediumType,
    HeatMediumType,
    OpticalMediumType,
)


class MultiPhysicsMedium(Tidy3dBaseModel):
    """
    Contains multiple multi-physical properties as defined for each solver medium.

    Examples
    --------
    For *silica* (:math:`SiO_2`):
        >>> import tidy3d as td
        >>> SiO2 = td.MultiPhysicsMedium(
        ...   optical=td.Medium(permittivity=3.9),
        ...   charge=td.ChargeInsulatorMedium(permittivity=3.9), # redefining permittivity
        ...   name="SiO2",
        ... )

    For a silicon ``MultiPhysicsMedium`` composed of an optical model
    from the material library and custom charge :class:`SemiconductorMedium`:
        >>> import tidy3d as td
        >>> default_multiphysics_Si = td.MultiPhysicsMedium(
        ...     optical=td.material_library['cSi']['Green2008'],
        ...     charge=td.SemiconductorMedium(
        ...         N_c=2.86e19,
        ...         N_v=3.1e19,
        ...         E_g=1.11,
        ...         mobility_n=td.CaugheyThomasMobility(
        ...             mu_min=52.2,
        ...             mu=1471.0,
        ...             ref_N=9.68e16,
        ...             exp_N=0.68,
        ...             exp_1=-0.57,
        ...             exp_2=-2.33,
        ...             exp_3=2.4,
        ...             exp_4=-0.146,
        ...         ),
        ...         mobility_p=td.CaugheyThomasMobility(
        ...             mu_min=44.9,
        ...             mu=470.5,
        ...             ref_N=2.23e17,
        ...             exp_N=0.719,
        ...             exp_1=-0.57,
        ...             exp_2=-2.33,
        ...             exp_3=2.4,
        ...             exp_4=-0.146,
        ...         ),
        ...         R=[
        ...             td.ShockleyReedHallRecombination(
        ...                 tau_n=3.3e-6,
        ...                 tau_p=4e-6
        ...             ),
        ...             td.RadiativeRecombination(
        ...                 r_const=1.6e-14
        ...             ),
        ...             td.AugerRecombination(
        ...                 c_n=2.8e-31,
        ...                 c_p=9.9e-32
        ...             ),
        ...         ],
        ...         delta_E_g=td.SlotboomBandGapNarrowing(
        ...             v1=6.92 * 1e-3,
        ...             n2=1.3e17,
        ...             c2=0.5,
        ...             min_N=1e15,
        ...         ),
        ...         N_a=0,
        ...         N_d=0
        ...     )
        ... )
    """

    name: Optional[str] = pd.Field(None, title="Name", description="Medium name")

    optical: Optional[OpticalMediumType] = pd.Field(
        None, title="Optical properties", description="Specifies optical properties."
    )

    # electrical: Optional[ElectricalMediumType] = pd.Field(
    #     None,
    #     title="Electrical properties",
    #     description="Specifies electrical properties for RF simulations. This is currently not in use.",
    # )

    heat: Optional[HeatMediumType] = pd.Field(
        None, title="Heat properties", description="Specifies properties for Heat simulations."
    )

    charge: Optional[ChargeMediumType] = pd.Field(
        None, title="Charge properties", description="Specifies properties for Charge simulations."
    )

    def __getattr__(self, name: str):
        """
        Delegate attribute lookup to inner media or fail fast.

        Parameters
        ----------
        name : str
            The attribute that could not be found on the ``MultiPhysicsMedium`` itself.

        Returns
        -------
        Any
            * The attribute value obtained from a delegated sub-medium when
            ``name`` is listed in ``DELEGATED_ATTRIBUTES``.
            * ``None`` when ``name`` is explicitly ignored (e.g. ``"__deepcopy__"``).

        Raises
        ------
        ValueError
            If ``name`` is neither ignored nor in the delegation map, signalling that
            the caller may have intended to access ``optical``, ``heat``, or
            ``charge`` directly.

        Notes
        -----
        Only the attributes enumerated in the local ``DELEGATED_ATTRIBUTES`` dict are
        forwarded.
        Extend that mapping as additional cross-medium shim behaviour becomes
        necessary.
        """
        # first check whether the attribute is already present
        try:
            return super().__getattr__(name)
        except AttributeError:
            pass

        IGNORED_ATTRIBUTES = ["__deepcopy__"]
        if name in IGNORED_ATTRIBUTES:
            return None

        DELEGATED_ATTRIBUTES = {
            "is_pmc": self.optical,
            "_eps_plot": self.optical,
            "viz_spec": self.optical,
            "eps_diagonal_numerical": self.optical,
            "eps_complex_to_nk": self.optical,
            "nonlinear_spec": self.optical,
            "is_pec": self.optical,
            "is_time_modulated": self.optical,
            "is_nonlinear": self.optical,
            "is_fully_anisotropic": self.optical,
            "is_custom": self.optical,
            "is_isotropic": self.optical,
            "is_spatially_uniform": self.optical,
            "_incompatible_material_types": self.optical,
            "frequency_range": self.optical,
            "eps_model": self.optical,
            "n_cfl": self.optical,
            "allow_gain": self.optical,
        }

        if name == "_has_incompatibilities":
            return (self.optical and self.optical._has_incompatibilities) or (
                self.charge and self.charge._has_incompatibilities
            )

        if name in DELEGATED_ATTRIBUTES:
            sub = DELEGATED_ATTRIBUTES[name]
            if sub is None:
                raise AttributeError(
                    f"Requested attribute {name!r}, but the optical medium is 'None' "
                    " on this 'MultiPhysicsMedium' instance."
                )
            return getattr(sub, name)

        raise AttributeError(
            f"MultiPhysicsMedium has no attribute called {name}. "
            "Did you mean to access the attribute of one of the optical, heat or charge media?"
        )

    @property
    def heat_spec(self):
        if self.heat is not None:
            return self.heat

        if self.optical is not None:
            return self.optical.heat_spec
        return None



"""
Tool for generating an S matrix automatically from a Tidy3d simulation and lumped port definitions.
"""

from __future__ import annotations

import numpy as np

from tidy3d.plugins.smatrix.data.data_array import ModalPortDataArray
from tidy3d.plugins.smatrix.data.modal import ModalComponentModelerData


def modal_construct_smatrix(modeler_data: ModalComponentModelerData) -> ModalPortDataArray:
    """Constructs the S-matrix from the data of a :class:`.ModalComponentModeler`.

    This function post-processes the :class:`.SimulationData` from a series of
    simulations to compute the scattering matrix (S-matrix).

    Parameters
    ----------
    modeler_data : ModalComponentModelerData
        The data from the :class:`.ModalComponentModeler` run, containing
        the modeler and the simulation data.

    Returns
    -------
    ModalPortDataArray
        The computed S-matrix.
    """

    max_mode_index_out, max_mode_index_in = modeler_data.modeler.max_mode_index
    num_modes_out = max_mode_index_out + 1
    num_modes_in = max_mode_index_in + 1
    port_names_out, port_names_in = modeler_data.modeler.port_names

    values = np.zeros(
        (
            len(port_names_out),
            len(port_names_in),
            num_modes_out,
            num_modes_in,
            len(modeler_data.modeler.freqs),
        ),
        dtype=complex,
    )
    coords = {
        "port_out": port_names_out,
        "port_in": port_names_in,
        "mode_index_out": range(num_modes_out),
        "mode_index_in": range(num_modes_in),
        "f": np.array(modeler_data.modeler.freqs),
    }
    s_matrix = ModalPortDataArray(values, coords=coords)

    # loop through source ports
    for col_index in modeler_data.modeler.matrix_indices_run_sim:
        port_name_in, mode_index_in = col_index
        port_in = modeler_data.modeler.get_port_by_name(port_name=port_name_in)

        sim_data = modeler_data.data[
            modeler_data.modeler.get_task_name(port=port_in, mode_index=mode_index_in)
        ]

        for row_index in modeler_data.modeler.matrix_indices_monitor:
            port_name_out, mode_index_out = row_index
            port_out = modeler_data.modeler.get_port_by_name(port_name=port_name_out)

            # directly compute the element
            mode_amps_data = sim_data[port_out.name].copy().amps
            dir_out = "-" if port_out.direction == "+" else "+"
            amp = mode_amps_data.sel(f=coords["f"], direction=dir_out, mode_index=mode_index_out)
            source_norm = modeler_data.modeler._normalization_factor(port_in, sim_data)
            s_matrix_elements = np.array(amp.data) / np.array(source_norm)
            coords_set = {
                "port_in": port_name_in,
                "mode_index_in": mode_index_in,
                "port_out": port_name_out,
                "mode_index_out": mode_index_out,
            }

            s_matrix = s_matrix._with_updated_data(data=s_matrix_elements, coords=coords_set)

    # element can be determined by user-defined mapping
    for (row_in, col_in), (row_out, col_out), mult_by in modeler_data.modeler.element_mappings:
        port_out_from, mode_index_out_from = row_in
        port_in_from, mode_index_in_from = col_in
        coords_from = {
            "port_in": port_in_from,
            "mode_index_in": mode_index_in_from,
            "port_out": port_out_from,
            "mode_index_out": mode_index_out_from,
        }

        port_out_to, mode_index_out_to = row_out
        port_in_to, mode_index_in_to = col_out
        elements_from = mult_by * s_matrix.loc[coords_from].values
        coords_to = {
            "port_in": port_in_to,
            "mode_index_in": mode_index_in_to,
            "port_out": port_out_to,
            "mode_index_out": mode_index_out_to,
        }
        s_matrix = s_matrix._with_updated_data(data=elements_from, coords=coords_to)

    return s_matrix



"""Utilities for converting between tidy3d versions."""

from __future__ import annotations

import functools
import json
from typing import Callable, Optional

import pydantic.v1 as pd
import yaml

from .components.base import Tidy3dBaseModel
from .exceptions import FileError, SetupError
from .log import log
from .version import __version__

"""Storing version numbers."""


class Version(pd.BaseModel):
    """Stores a version number (excluding patch)."""

    major: int
    minor: int

    @classmethod
    def from_string(cls, string: Optional[str] = None) -> Version:
        """Return Version from a version string."""
        if string is None:
            return cls.from_string(string=__version__)

        try:
            version_numbers = string.split(".")
            version = cls(major=version_numbers[0], minor=version_numbers[1])
        except Exception as e:
            raise SetupError(f"version string {string} can't be parsed.") from e
        return version

    @property
    def as_tuple(self):
        """version as a tuple, leave out patch for now."""
        return (self.major, self.minor)

    def __hash__(self):
        """define a hash."""
        return hash(self.as_tuple)

    def __str__(self):
        """Convert back to string."""
        return f"{self.major}.{self.minor}"

    def __eq__(self, other):
        """versions equal."""
        return (self.major == other.major) and (self.minor == other.minor)

    def __lt__(self, other):
        """self < other."""
        if self.major < other.major:
            return True
        if self.major == other.major:
            return self.minor < other.minor
        return False

    def __gt__(self, other):
        """self > other."""
        if self.major > other.major:
            return True
        if self.major == other.major:
            return self.minor > other.minor
        return False

    def __le__(self, other):
        """self <= other."""
        return (self < other) or (self == other)

    def __ge__(self, other):
        """self >= other."""
        return (self > other) or (self == other)


CurrentVersion = Version.from_string(__version__)

"""Class for updating simulation objects."""


class Updater(pd.BaseModel):
    """Converts a tidy3d simulation.json file to an up-to-date Simulation instance."""

    sim_dict: dict

    @classmethod
    def from_file(cls, fname: str) -> Updater:
        """Dictionary representing the simulation loaded from file."""

        # TODO: fix this, it broke
        if any(ext in fname for ext in (".hdf5", ".gz")):
            sim_dict = Tidy3dBaseModel.from_file(fname=fname).dict()

        else:
            # try:
            with open(fname, encoding="utf-8") as f:
                if ".json" in fname:
                    sim_dict = json.load(f)
                elif ".yaml" in fname:
                    sim_dict = yaml.safe_load(f)
                else:
                    raise FileError('file extension must be ".json", ".yaml", ".hdf5", or ".gz"')

            # except Exception as e:
            #     raise FileError(f"Could not load file {fname}") from e

        return cls(sim_dict=sim_dict)

    @classmethod
    def from_string(cls, sim_dict_str: str) -> Updater:
        """Dictionary representing the simulation loaded from string."""
        sim_dict = json.loads(sim_dict_str)
        return cls(sim_dict=sim_dict)

    @property
    def version(self) -> Version:
        """Version of the supplied file."""
        version_string = self.sim_dict.get("version")
        if version_string is None:
            raise SetupError("Could not find a version in the supplied json.")
        return Version.from_string(version_string)

    def get_update_function(self):
        """Get the highest update version <= self.version."""
        leq_versions = [v for v in UPDATE_MAP if v <= self.version]
        if not leq_versions:
            raise SetupError(f"An update version <= {self.version} not found in update map.")
        update_version = max(leq_versions)
        return UPDATE_MAP[update_version]

    def get_next_version(self) -> Version:
        """Get the next version after self.version."""
        gt_versions = [v for v in UPDATE_MAP if v > self.version]
        if not gt_versions:
            return CurrentVersion
        return str(min(gt_versions))

    def update_to_current(self) -> dict:
        """Update supplied simulation dictionary to current version."""
        if self.version == CurrentVersion:
            self.sim_dict["version"] = __version__
            return self.sim_dict
        log.warning(f"updating Simulation from {self.version} to {CurrentVersion}")
        while self.version != CurrentVersion:
            update_fn = self.get_update_function()
            self.sim_dict = update_fn(self.sim_dict)
            self.sim_dict["version"] = str(self.get_next_version())
        self.sim_dict["version"] = __version__
        return self.sim_dict

    def __eq__(self, other: Updater) -> bool:
        """Is Updater equal to another one?"""
        return self.sim_dict == other.sim_dict


"""Update conversion functions."""

# versions will be dynamically mapped in this table when the update functions are initialized.
UPDATE_MAP = {}


def updates_from_version(version_from_string: str):
    """Decorates a sim_dict update function to change the version."""

    # make sure the version strings are legit
    from_version = Version.from_string(version_from_string)

    def decorator(update_fn):
        """The actual decorator that gets returned by `updates_to_version('x.y.z')`"""

        @functools.wraps(update_fn)
        def new_update_function(sim_dict: dict) -> dict:
            """Update function that automatically adds version string."""

            return update_fn(sim_dict)

        UPDATE_MAP[from_version] = new_update_function

        return new_update_function

    return decorator


def iterate_update_dict(update_dict: dict, update_types: dict[str, Callable]):
    """Recursively iterate nested ``update_dict``. For any nested ``nested_dict`` found,
    apply an update function if its ``nested_dict["type"]`` is in the keys of the ``update_types``
    dictionary. Also iterates lists and tuples.
    """

    if isinstance(update_dict, dict):
        # Update if we need to, and iterate recursively all items
        if update_dict.get("type") in update_types:
            update_types[update_dict["type"]](update_dict)
        for item in update_dict.values():
            iterate_update_dict(item, update_types)
    elif isinstance(update_dict, (list, tuple)):
        # Try other iterables
        for item in update_dict:
            iterate_update_dict(item, update_types)


@updates_from_version("1.8")
def update_1_8(sim_dict: dict) -> dict:
    """Updates version 1.8."""

    def fix_missing_scalar_field(mnt_dict: dict) -> dict:
        for key, val in mnt_dict["field_dataset"].items():
            if isinstance(val, str) and val == "XR.DATAARRAY":
                mnt_dict["field_dataset"][key] = "ScalarFieldDataArray"
        return mnt_dict

    iterate_update_dict(
        update_dict=sim_dict,
        update_types={
            "CustomFieldSource": fix_missing_scalar_field,
        },
    )
    return sim_dict


@updates_from_version("1.7")
def update_1_7(sim_dict: dict) -> dict:
    """Updates version 1.7."""

    def fix_angle_info(mnt_dict: dict) -> dict:
        mnt_dict["type"] = "FieldProjectionAngleMonitor"
        mnt_dict.pop("fields")
        mnt_dict.pop("medium")
        mnt_dict["proj_distance"] = 1e6
        return mnt_dict

    def fix_cartesian_info(mnt_dict: dict) -> dict:
        mnt_dict["type"] = "FieldProjectionCartesianMonitor"
        mnt_dict.pop("fields")
        mnt_dict.pop("medium")
        dist = mnt_dict.pop("plane_distance")
        mnt_dict["proj_distance"] = dist
        axis = mnt_dict.pop("plane_axis")
        mnt_dict["proj_axis"] = axis
        return mnt_dict

    def fix_kspace_info(mnt_dict: dict) -> dict:
        mnt_dict["type"] = "FieldProjectionKSpaceMonitor"
        mnt_dict.pop("fields")
        mnt_dict.pop("medium")
        mnt_dict["proj_distance"] = 1e6
        axis = mnt_dict.pop("u_axis")
        mnt_dict["proj_axis"] = axis
        return mnt_dict

    def fix_diffraction_info(mnt_dict: dict) -> dict:
        mnt_dict.pop("medium", None)
        mnt_dict.pop("orders_x", None)
        mnt_dict.pop("orders_y", None)
        return mnt_dict

    def fix_bloch_vec(mnt_dict: dict) -> dict:
        mnt_dict["bloch_vec"] = mnt_dict["bloch_vec"]["real"]
        return mnt_dict

    iterate_update_dict(
        update_dict=sim_dict,
        update_types={
            "Near2FarAngleMonitor": fix_angle_info,
            "Near2FarCartesianMonitor": fix_cartesian_info,
            "Near2FarKSpaceMonitor": fix_kspace_info,
            "DiffractionMonitor": fix_diffraction_info,
            "BlochBoundary": fix_bloch_vec,
        },
    )
    return sim_dict


@updates_from_version("1.6")
def update_1_6(sim_dict: dict) -> dict:
    """Updates version 1.6."""
    if "grid_size" in sim_dict:
        sim_dict.pop("grid_size")
    return sim_dict


@updates_from_version("1.5")
def update_1_5(sim_dict: dict) -> dict:
    """Updates version 1.5."""

    def fix_mode_field_mnt(mnt_dict: dict) -> dict:
        mnt_dict["type"] = "ModeSolverMonitor"
        return mnt_dict

    iterate_update_dict(update_dict=sim_dict, update_types={"ModeFieldMonitor": fix_mode_field_mnt})
    return sim_dict


@updates_from_version("1.4")
def update_1_4(sim_dict: dict) -> dict:
    """Updates version 1.4."""

    def fix_polyslab(geo_dict):
        """Fix a PolySlab dictionary."""
        geo_dict.pop("length", None)
        geo_dict.pop("center", None)

    def fix_modespec(ms_dict):
        """Fix a ModeSpec dictionary."""
        sort_by = ms_dict.pop("sort_by", None)
        if sort_by and sort_by != "largest_neff":
            log.warning(
                "ModeSpec.sort_by was removed in Tidy3D 1.5.0, reverting to sorting by "
                "largest effective index. Use ModeSpec.filter_pol to select polarization instead."
            )

    def fix_geometry_group(geo_dict):
        """Fix a GeometryGroup dictionary."""
        geo_dict.pop("center", None)

    update_types = {
        "PolySlab": fix_polyslab,
        "ModeSpec": fix_modespec,
        "GeometryGroup": fix_geometry_group,
    }

    iterate_update_dict(update_dict=sim_dict, update_types=update_types)

    return sim_dict


@updates_from_version("1.3")
def update_1_3(sim_dict: dict) -> dict:
    """Updates version 1.3."""

    sim_dict["boundary_spec"] = {"x": {}, "y": {}, "z": {}}
    for dim, pml_layer in zip(["x", "y", "z"], sim_dict["pml_layers"]):
        sim_dict["boundary_spec"][dim]["plus"] = pml_layer
        sim_dict["boundary_spec"][dim]["minus"] = pml_layer
    sim_dict.pop("pml_layers")
    return sim_dict



from __future__ import annotations

from typing import Union

from tidy3d.components.data.data_array import (
    CurrentFreqDataArray,
    CurrentFreqModeDataArray,
    VoltageFreqDataArray,
    VoltageFreqModeDataArray,
)
from tidy3d.plugins.smatrix.ports.coaxial_lumped import CoaxialLumpedPort
from tidy3d.plugins.smatrix.ports.rectangular_lumped import LumpedPort
from tidy3d.plugins.smatrix.ports.wave import WavePort

LumpedPortType = Union[LumpedPort, CoaxialLumpedPort]
TerminalPortType = Union[LumpedPortType, WavePort]
PortVoltageType = Union[VoltageFreqDataArray, VoltageFreqModeDataArray]
PortCurrentType = Union[CurrentFreqDataArray, CurrentFreqModeDataArray]



"""Abstract bases for classes that define how data is recorded from simulation."""

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.base import Box
from tidy3d.components.types import ArrayFloat1D, Axis, Numpy
from tidy3d.components.validators import _warn_unsupported_traced_argument
from tidy3d.components.viz import PlotParams, plot_params_monitor


class AbstractMonitor(Box, ABC):
    """Abstract base class for steady-state monitors."""

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for monitor.",
        min_length=1,
    )

    _warn_traced_center = _warn_unsupported_traced_argument("center")
    _warn_traced_size = _warn_unsupported_traced_argument("size")

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a Monitor object."""
        return plot_params_monitor

    @cached_property
    def geometry(self) -> Box:
        """:class:`Box` representation of monitor.

        Returns
        -------
        :class:`Box`
            Representation of the monitor geometry as a :class:`Box`.
        """
        return Box(center=self.center, size=self.size)

    @abstractmethod
    def storage_size(self, num_cells: int, tmesh: ArrayFloat1D) -> int:
        """Size of monitor storage given the number of points after discretization.

        Parameters
        ----------
        num_cells : int
            Number of grid cells within the monitor after discretization by a :class:`.Simulation`.
        tmesh : Array
            The discretized time mesh of a :class:`.Simulation`.

        Returns
        -------
        int
            Number of bytes to be stored in monitor.
        """

    def downsample(self, arr: Numpy, axis: Axis) -> Numpy:
        """Downsample a 1D array making sure to keep the first and last entries, based on the
        spatial interval defined for the ``axis``.

        Parameters
        ----------
        arr : Numpy
            A 1D array of arbitrary type.
        axis : Axis
            Axis for which to select the interval_space defined for the monitor.

        Returns
        -------
        Numpy
            Downsampled array.
        """

        size = len(arr)
        interval = self.interval_space[axis]
        # There should always be at least 3 indices for "surface" monitors. Also, if the
        # size along this dim is already smaller than the interval, then don't downsample.
        if size < 4 or (size - 1) <= interval:
            return arr
        # make sure the last index is always included
        inds = np.arange(0, size, interval)
        if inds[-1] != size - 1:
            inds = np.append(inds, size - 1)
        return arr[inds]

    def downsampled_num_cells(self, num_cells: tuple[int, int, int]) -> tuple[int, int, int]:
        """Given a tuple of the number of cells spanned by the monitor along each dimension,
        return the number of cells one would have after downsampling based on ``interval_space``.
        """
        arrs = [np.arange(ncells) for ncells in num_cells]
        return tuple((self.downsample(arr, axis=dim).size for dim, arr in enumerate(arrs)))



"""Exports all data types from the .base module for easy access."""

from __future__ import annotations

from tidy3d.components.types.base import (
    TYPE_TAG_STR,
    ArrayComplex1D,
    ArrayComplex2D,
    ArrayComplex3D,
    ArrayComplex4D,
    ArrayFloat1D,
    ArrayFloat2D,
    ArrayFloat3D,
    ArrayFloat4D,
    ArrayInt1D,
    ArrayLike,
    AuxField,
    Ax,
    Axis,
    Axis2D,
    Bound,
    BoxSurface,
    ClipOperationType,
    ColormapType,
    Complex,
    ComplexNumber,
    Coordinate,
    Coordinate2D,
    CoordinateOptional,
    Direction,
    EMField,
    EpsSpecType,
    FieldType,
    FieldVal,
    FreqArray,
    FreqBound,
    FreqBoundMax,
    FreqBoundMin,
    GridSize,
    InterpMethod,
    LengthUnit,
    LumpDistType,
    MatrixReal4x4,
    ModeSolverType,
    Numpy,
    ObsGridArray,
    PermittivityComponent,
    PlanePosition,
    PlotScale,
    PlotVal,
    Polarization,
    PolarizationBasis,
    PoleAndResidue,
    PriorityMode,
    RealFieldVal,
    ScalarSymmetry,
    Shapely,
    Size,
    Size1D,
    Symmetry,
    TensorReal,
    TrackFreq,
    Undefined,
    UnitsZBF,
    annotate_type,
    constrained_array,
    tidycomplex,
    xyz,
)
from tidy3d.components.types.third_party import TrimeshType
from tidy3d.components.types.utils import _add_schema

__all__ = [
    "TYPE_TAG_STR",
    "ArrayComplex1D",
    "ArrayComplex2D",
    "ArrayComplex3D",
    "ArrayComplex4D",
    "ArrayFloat1D",
    "ArrayFloat2D",
    "ArrayFloat3D",
    "ArrayFloat4D",
    "ArrayInt1D",
    "ArrayLike",
    "AuxField",
    "Ax",
    "Axis",
    "Axis2D",
    "Bound",
    "BoxSurface",
    "ClipOperationType",
    "ColormapType",
    "Complex",
    "ComplexNumber",
    "Coordinate",
    "Coordinate2D",
    "CoordinateOptional",
    "Direction",
    "EMField",
    "EpsSpecType",
    "FieldType",
    "FieldVal",
    "FreqArray",
    "FreqBound",
    "FreqBoundMax",
    "FreqBoundMin",
    "GridSize",
    "InterpMethod",
    "LengthUnit",
    "LumpDistType",
    "MatrixReal4x4",
    "ModeSolverType",
    "Numpy",
    "ObsGridArray",
    "PermittivityComponent",
    "PlanePosition",
    "PlotScale",
    "PlotVal",
    "Polarization",
    "PolarizationBasis",
    "PoleAndResidue",
    "PriorityMode",
    "RealFieldVal",
    "ScalarSymmetry",
    "Shapely",
    "Size",
    "Size1D",
    "Symmetry",
    "TensorReal",
    "TrackFreq",
    "TrimeshType",
    "Undefined",
    "UnitsZBF",
    "_add_schema",
    "annotate_type",
    "constrained_array",
    "tidycomplex",
    "xyz",
]



"""Default values for KLayout DRC."""

from __future__ import annotations

from pathlib import Path

DEFAULT_GDSFILE = Path("layout.gds")
DEFAULT_RESULTSFILE = Path("drc_results.lyrdb")
DEFAULT_VERBOSE = True



from __future__ import annotations

from typing import Callable, Optional, Union

import autograd.numpy as np
import pydantic.v1 as pd
from numpy.typing import NDArray

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import ArrayFloat2D
from tidy3d.plugins.autograd.types import PaddingType

from .parametrizations import FilterAndProject


class ErosionDilationPenalty(Tidy3dBaseModel):
    """A class that computes a penalty for erosion/dilation of a parameter map not being unity."""

    radius: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Radius", description="The radius of the kernel."
    )
    dl: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Grid Spacing", description="The grid spacing."
    )
    size_px: Union[int, tuple[int, ...]] = pd.Field(
        None, title="Size in Pixels", description="The size of the kernel in pixels."
    )
    beta: pd.NonNegativeFloat = pd.Field(
        20.0, title="Beta", description="The beta parameter for the tanh projection."
    )
    eta: pd.NonNegativeFloat = pd.Field(
        0.5, title="Eta", description="The eta parameter for the tanh projection."
    )
    filter_type: str = pd.Field(
        "conic", title="Filter Type", description="The type of filter to create."
    )
    padding: PaddingType = pd.Field(
        "reflect", title="Padding", description="The padding mode to use."
    )
    delta_eta: float = pd.Field(
        0.01,
        title="Delta Eta",
        description="The binarization threshold for erosion and dilation operations.",
    )

    def __call__(self, array: NDArray) -> float:
        """Compute the erosion/dilation penalty for a given array.

        Parameters
        ----------
        array : np.ndarray
            The input array to compute the penalty for.

        Returns
        -------
        float
            The computed erosion/dilation penalty.
        """
        filtproj = FilterAndProject(
            radius=self.radius,
            dl=self.dl,
            size_px=self.size_px,
            beta=self.beta,
            eta=self.eta,
            filter_type=self.filter_type,
            padding=self.padding,
        )

        eta_dilate = 0.0 + self.delta_eta
        eta_eroded = 1.0 - self.delta_eta

        def _dilate(arr: NDArray):
            return filtproj(arr, eta=eta_dilate)

        def _erode(arr: NDArray):
            return filtproj(arr, eta=eta_eroded)

        def _open(arr: NDArray):
            return _dilate(_erode(arr))

        def _close(arr: NDArray):
            return _erode(_dilate(arr))

        diff = _close(array) - _open(array)

        if not np.any(diff):
            return 0.0

        return np.linalg.norm(diff) / np.sqrt(diff.size)


def make_erosion_dilation_penalty(
    radius: Union[float, tuple[float, ...]],
    dl: Union[float, tuple[float, ...]],
    *,
    size_px: Optional[Union[int, tuple[int, ...]]] = None,
    beta: float = 20.0,
    eta: float = 0.5,
    delta_eta: float = 0.01,
    padding: PaddingType = "reflect",
) -> Callable:
    """Computes a penalty for erosion/dilation of a parameter map not being unity.

    See Also
    --------
    :func:`~penalties.ErosionDilationPenalty`.
    """
    return ErosionDilationPenalty(
        radius=radius,
        dl=dl,
        size_px=size_px,
        beta=beta,
        eta=eta,
        delta_eta=delta_eta,
        padding=padding,
    )


def curvature(dp: NDArray, ddp: NDArray) -> NDArray:
    """Calculate the curvature at a point given the first and second derivatives.

    Parameters
    ----------
    dp : np.ndarray
        The first derivative at the point, with (x, y) entries in the first dimension.
    ddp : np.ndarray
        The second derivative at the point, with (x, y) entries in the first dimension.

    Returns
    -------
    np.ndarray
        The curvature at the given point.

    Notes
    -----
    The curvature can be positive or negative, indicating the direction of curvature.
    The radius of curvature is defined as 1 / |κ|, where κ is the curvature.
    """
    num = dp[0] * ddp[1] - dp[1] * ddp[0]
    den = np.power(dp[0] ** 2 + dp[1] ** 2, 1.5)
    return num / den


def bezier_with_grads(
    t: float, p0: NDArray, pc: NDArray, p2: NDArray
) -> tuple[NDArray, NDArray, NDArray]:
    """Compute the Bezier curve and its first and second derivatives at a given point.

    Parameters
    ----------
    t : float
        The parameter at which to evaluate the Bezier curve.
    p0 : np.ndarray
        The first control point of the Bezier curve.
    pc : np.ndarray
        The central control point of the Bezier curve.
    p2 : np.ndarray
        The last control point of the Bezier curve.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple containing the Bezier curve value, its first derivative, and its second derivative at the given point.
    """
    p1 = 2 * pc - p0 / 2 - p2 / 2
    b = (1 - t) ** 2 * (p0 - p1) + p1 + t**2 * (p2 - p1)
    dbdt = 2 * ((1 - t) * (p1 - p0) + t * (p2 - p1))
    dbd2t = 2 * (p0 - 2 * p1 + p2)
    return b, dbdt, dbd2t


def bezier_curvature(x: NDArray, y: NDArray, t: Union[NDArray, float] = 0.5) -> NDArray:
    """
    Calculate the curvature of a Bezier curve at a given parameter t.

    Parameters
    ----------
    x : np.ndarray
        The x-coordinates of the control points.
    y : np.ndarray
        The y-coordinates of the control points.
    t : Union[np.ndarray, float] = 0.5
        The parameter at which to evaluate the curvature.

    Returns
    -------
    np.ndarray
        The curvature of the Bezier curve at the given parameter t.
    """
    p = np.stack((x, y), axis=1)
    _, dbdt, dbd2t = bezier_with_grads(t, p[:-2], p[1:-1], p[2:])
    return curvature(dbdt.T, dbd2t.T)


def make_curvature_penalty(
    min_radius: float, alpha: float = 1.0, kappa: float = 10.0, *, eps: float = 1e-6
) -> Callable:
    """Create a penalty function based on the curvature of a set of points.

    Parameters
    ----------
    min_radius : float
        The minimum radius of curvature.
    alpha : float = 1.0
        Scaling factor for the penalty.
    kappa : float = 10.0
        Exponential factor for the penalty.
    eps : float = 1e-6
        A small value to avoid division by zero.

    Returns
    -------
    Callable
        A function that computes the curvature penalty for a given set of points.

    Notes
    -----
    The penalty function is defined as:

    .. math::

        p(r) = \\frac{\\mathrm{exp}(-\\kappa(r - r_{min}))}{1 + \\mathrm{exp}(-\\kappa(r - r_{min}))}

    This formula was described by A. Micheals et al.
    "Leveraging continuous material averaging for inverse electromagnetic design",
    Optics Express (2018).
    """

    def _curvature_penalty(points: ArrayFloat2D) -> float:
        xs, ys = np.array(points).T
        crv = bezier_curvature(xs, ys)
        curvature_radius = 1 / (np.abs(crv) + eps)
        arg = kappa * (curvature_radius - min_radius)
        exp_arg = np.exp(-arg)
        penalty = alpha * (exp_arg / (1 + exp_arg))
        return np.mean(penalty)

    return _curvature_penalty



"""
This module contains the CLI commands for testing the tidy3d package. This includes testing the base package and the
notebooks in order to achieve reproducibility between hardwares.
"""

from __future__ import annotations

import click

from .index import develop
from .install import install_in_poetry
from .utils import echo_and_run_subprocess

__all__ = [
    "test_in_environment_command",
    "test_options",
]


def test_options(options: list):
    """
    Inclusive rather than exclusive tests in a given set of environments.

    Parameters
    ----------
    options : list
        A list of options for which tests to run. Options are 'base' and 'notebooks'.
    """
    if "base" in options:
        echo_and_run_subprocess(["poetry", "run", "pytest", "-rA", "tests"])
    if "notebooks" in options:
        echo_and_run_subprocess(["poetry", "run", "pytest", "-rA", "tests/full_test_notebooks.py"])


@click.option(
    "--types",
    default=["base"],
    help="Types of tests to run. Defaults to 'base'. Other options",
    type=list,
)
@click.option(
    "--env",
    default="dev",
    help="Poetry environment to install. Defaults to 'dev'.",
    type=str,
)
@develop.command(
    name="test-in-envrionment", help="Installs the specified poetry environment and tests"
)
def test_in_environment_command(types: list, env: str = "dev"):
    """
    Installs a poetry environment specified by the extra definition in pyproject.toml and runs tests with pytest and
    any additional arguments. Requires a poetry installation so make sure to verify the installation previously.

    If the environment is already installed, it will be reinstalled to ensure the latest version of a reproducible
    envrionment is used.

    Parameters
    ----------
    types : list
        A list of options for which tests to run.
    env : str
        The name of the poetry environment to install. Defaults to 'dev'. See pyproject.toml
    """
    install_in_poetry(env)
    test_options(types)



"""Computes approximate parameters for a pair of coupled microstrips using the model given in Kirschning and Jansen [1].

References
----------
[1]     Kirschning, M., & Jansen, R. H. (1984). Accurate Wide-Range Design Equations for
        the Frequency-Dependent Characteristic of Parallel Coupled Microstrip Lines. IEEE
        Transactions on Microwave Theory and Techniques, 32(1), 83-90.
"""

from __future__ import annotations

import numpy as np

from . import microstrip


def _epsilon_e_even(relative_permittivity: float, width: float, height: float, gap: float) -> float:
    """Equation 3 [1] - Reuses functionality from the ``microstrip`` module."""
    normalized_width = width / height
    normalized_gap = gap / height
    u = normalized_width
    g = normalized_gap
    er = relative_permittivity
    v = u * (20 + g**2) / (10 + g**2) + g * np.exp(-g)
    a_v = microstrip._a(v)
    b_er = microstrip._b(er)
    scale_factor = (1 + 10 / v) ** (-a_v * b_er)
    return (er + 1) / 2 + (er - 1) / 2 * scale_factor


def _epsilon_e_odd(
    relative_permittivity: float, width: float, height: float, gap: float, e_eff: float
) -> float:
    """Equation 4 [1]"""
    normalized_width = width / height
    normalized_gap = gap / height
    u = normalized_width
    g = normalized_gap
    er = relative_permittivity
    # e_eff = MicrostripModel._epsilon_e(normalized_width, relative_permittivity)
    a_o = 0.7287 * (e_eff - 0.5 * (er + 1)) * (1 - np.exp(-0.179 * u))
    b_o = 0.747 * er / (0.15 + er)
    c_o = b_o - (b_o - 0.207) * np.exp(-0.414 * u)
    d_o = 0.593 + 0.694 * np.exp(-0.562 * u)
    return (0.5 * (er + 1) + a_o - e_eff) * np.exp(-c_o * g**d_o) + e_eff


def _z0_even_odd(
    width: float,
    height: float,
    gap: float,
    e_eff_even: float,
    e_eff_odd: float,
    z0: float,
    e_eff: float,
) -> tuple[float, float]:
    """Computes the characteristic impedance of the even and odd modes for coupled microstrip lines.
    Equations 8 and 9 [1]"""
    normalized_width = width / height
    normalized_gap = gap / height
    u = normalized_width
    g = normalized_gap
    Q1 = 0.8695 * u**0.194
    Q2 = 1 + 0.7519 * g + 0.189 * g**2.31
    Q3 = 0.1975 + (16.6 + (8.4 / g) ** 6) ** -0.387 + np.log(g**10 / (1 + (g / 3.4) ** 10)) / 241
    Q4 = (2 * Q1 / Q2) * (np.exp(-g) * u**Q3 + (2 - np.exp(-g)) * u**-Q3) ** -1
    Q5 = 1.794 + 1.14 * np.log(1 + 0.638 / (g + 0.517 * g**2.43))
    Q6 = 0.2305 + np.log(g**10 / (1 + (g / 5.8) ** 10)) / 281.3 + np.log(1 + 0.598 * g**1.154) / 5.1
    Q7 = (10 + 190 * g**2) / (1 + 82.3 * g**3)
    Q8 = np.exp(-6.5 - 0.95 * np.log(g) - (g / 0.15) ** 5)
    Q9 = np.log(Q7) * (Q8 + 1 / 16.5)
    Q10 = Q2**-1 * (Q2 * Q4 - Q5 * np.exp(np.log(u) * Q6 * u**-Q9))
    z0_even = z0 * (e_eff / e_eff_even) ** 0.5 * 1 / (1 - (z0 / 377) * e_eff**0.5 * Q4)
    z0_odd = z0 * (e_eff / e_eff_odd) ** 0.5 * 1 / (1 - (z0 / 377) * e_eff**0.5 * Q10)
    return (z0_even, z0_odd)


def compute_line_params(
    relative_permittivity: float, width: float, height: float, gap: float
) -> tuple[float, float, float, float]:
    """Computes an approximation for the parameters of coupled microstrip lines
    assuming the quasistatic regime and 0 thickness [1].

    Parameters
    ----------
    relative_permittivity : float
        Relative electric permittivity of the substrate material.
    width : float
        Width of strips.
    height : float
        Distance between ground and bottom of strip.
    gap : float
        Spacing between the two microstrips.

    Returns
    -------
    ``tuple``
        ``Tuple`` containing the characteristic impedance and effective relative permittivity
        for the even and odd modes of the coupled lines."""

    (z0, e_eff) = microstrip.compute_line_params(relative_permittivity, width, height, 0)
    e_eff_even = _epsilon_e_even(relative_permittivity, width, height, gap)
    e_eff_odd = _epsilon_e_odd(relative_permittivity, width, height, gap, e_eff)
    (z0_even, z0_odd) = _z0_even_odd(width, height, gap, e_eff_even, e_eff_odd, z0, e_eff)
    return (z0_even, z0_odd, e_eff_even, e_eff_odd)



from __future__ import annotations

from functools import wraps
from typing import Optional

from tidy3d.components.types import Ax, Axis, LengthUnit
from tidy3d.constants import UnitScaling
from tidy3d.exceptions import Tidy3dKeyError


def _create_unit_aware_locator():
    """Create UnitAwareLocator lazily due to matplotlib import restrictions."""
    import matplotlib.ticker as ticker

    class UnitAwareLocator(ticker.Locator):
        """Custom tick locator that places ticks at nice positions in the target unit."""

        def __init__(self, scale_factor: float):
            """
            Parameters
            ----------
            scale_factor : float
                Factor to convert from micrometers to the target unit.
            """
            super().__init__()
            self.scale_factor = scale_factor

        def __call__(self):
            vmin, vmax = self.axis.get_view_interval()
            return self.tick_values(vmin, vmax)

        def view_limits(self, vmin, vmax):
            """Override to prevent matplotlib from adjusting our limits."""
            return vmin, vmax

        def tick_values(self, vmin, vmax):
            # convert the view range to the target unit
            vmin_unit = vmin * self.scale_factor
            vmax_unit = vmax * self.scale_factor

            # tolerance for floating point comparisons in target unit
            unit_range = vmax_unit - vmin_unit
            unit_tol = unit_range * 1e-8

            locator = ticker.MaxNLocator(nbins=11, prune=None, min_n_ticks=2)

            ticks_unit = locator.tick_values(vmin_unit, vmax_unit)

            # ensure we have ticks that cover the full range
            if len(ticks_unit) > 0:
                if ticks_unit[0] > vmin_unit + unit_tol or ticks_unit[-1] < vmax_unit - unit_tol:
                    # try with fewer bins to get better coverage
                    for n in [10, 9, 8, 7, 6, 5]:
                        locator = ticker.MaxNLocator(nbins=n, prune=None, min_n_ticks=2)
                        ticks_unit = locator.tick_values(vmin_unit, vmax_unit)
                        if (
                            len(ticks_unit) >= 3
                            and ticks_unit[0] <= vmin_unit + unit_tol
                            and ticks_unit[-1] >= vmax_unit - unit_tol
                        ):
                            break

                # if still no good coverage, manually ensure edge coverage
                if len(ticks_unit) > 0:
                    if (
                        ticks_unit[0] > vmin_unit + unit_tol
                        or ticks_unit[-1] < vmax_unit - unit_tol
                    ):
                        # find a reasonable step size from existing ticks
                        if len(ticks_unit) > 1:
                            step = ticks_unit[1] - ticks_unit[0]
                        else:
                            step = unit_range / 5

                        # extend the range to ensure coverage
                        extended_min = vmin_unit - step
                        extended_max = vmax_unit + step

                        # try one more time with extended range
                        locator = ticker.MaxNLocator(nbins=8, prune=None, min_n_ticks=2)
                        ticks_unit = locator.tick_values(extended_min, extended_max)

                        # filter to reasonable bounds around the original range
                        ticks_unit = [
                            t
                            for t in ticks_unit
                            if t >= vmin_unit - step / 2 and t <= vmax_unit + step / 2
                        ]

            # convert the nice ticks back to the original data unit (micrometers)
            ticks_um = ticks_unit / self.scale_factor

            # filter to ensure ticks are within bounds (with small tolerance)
            eps = (vmax - vmin) * 1e-8
            return [tick for tick in ticks_um if vmin - eps <= tick <= vmax + eps]

    return UnitAwareLocator


def make_ax() -> Ax:
    """makes an empty ``ax``."""
    import matplotlib.pyplot as plt

    _, ax = plt.subplots(1, 1, tight_layout=True)
    return ax


def add_ax_if_none(plot):
    """Decorates ``plot(*args, **kwargs, ax=None)`` function.
    if ax=None in the function call, creates an ax and feeds it to rest of function.
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function using a generated ax if None."""
        if kwargs.get("ax") is None:
            ax = make_ax()
            kwargs["ax"] = ax
        return plot(*args, **kwargs)

    return _plot


def equal_aspect(plot):
    """Decorates a plotting function returning a matplotlib axes.
    Ensures the aspect ratio of the returned axes is set to equal.
    Useful for 2D plots, like sim.plot() or sim_data.plot_fields()
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function with equal aspect ratio axes returned."""
        ax = plot(*args, **kwargs)
        ax.set_aspect("equal")
        return ax

    return _plot


def set_default_labels_and_title(
    axis_labels: tuple[str, str],
    axis: Axis,
    position: float,
    ax: Ax,
    plot_length_units: Optional[LengthUnit] = None,
) -> Ax:
    """Adds axis labels and title to plots involving spatial dimensions.
    When the ``plot_length_units`` are specified, the plot axes are scaled, and
    the title and axis labels include the desired units.
    """

    import matplotlib.ticker as ticker

    xlabel = axis_labels[0]
    ylabel = axis_labels[1]
    if plot_length_units is not None:
        if plot_length_units not in UnitScaling:
            raise Tidy3dKeyError(
                f"Provided units '{plot_length_units}' are not supported. "
                f"Please choose one of '{LengthUnit}'."
            )
        ax.set_xlabel(f"{xlabel} ({plot_length_units})")
        ax.set_ylabel(f"{ylabel} ({plot_length_units})")

        scale_factor = UnitScaling[plot_length_units]

        # for imperial units, use custom tick locator for nice tick positions
        if plot_length_units in ["mil", "in"]:
            UnitAwareLocator = _create_unit_aware_locator()
            x_locator = UnitAwareLocator(scale_factor)
            y_locator = UnitAwareLocator(scale_factor)
            ax.xaxis.set_major_locator(x_locator)
            ax.yaxis.set_major_locator(y_locator)

        formatter = ticker.FuncFormatter(lambda y, _: f"{y * scale_factor:.2f}")

        ax.xaxis.set_major_formatter(formatter)
        ax.yaxis.set_major_formatter(formatter)

        position_scaled = position * scale_factor
        ax.set_title(f"cross section at {'xyz'[axis]}={position_scaled:.2f} ({plot_length_units})")
    else:
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(f"cross section at {'xyz'[axis]}={position:.2f}")
    return ax



"""Web API for mode solver"""

from __future__ import annotations

import os
import pathlib
import tempfile
import time
from datetime import datetime
from typing import Callable, Literal, Optional, Union

import pydantic.v1 as pydantic
from botocore.exceptions import ClientError
from joblib import Parallel, delayed
from rich.progress import Progress

from tidy3d.components.data.monitor_data import ModeSolverData
from tidy3d.components.eme.simulation import EMESimulation
from tidy3d.components.medium import AbstractCustomMedium
from tidy3d.components.simulation import Simulation
from tidy3d.exceptions import SetupError, WebError
from tidy3d.log import get_logging_console, log
from tidy3d.plugins.mode.mode_solver import MODE_MONITOR_NAME, ModeSolver
from tidy3d.version import __version__
from tidy3d.web.core.core_config import get_logger_console
from tidy3d.web.core.environment import Env
from tidy3d.web.core.http_util import http
from tidy3d.web.core.s3utils import download_file, download_gz_file, upload_file
from tidy3d.web.core.task_core import Folder
from tidy3d.web.core.types import PayType, ResourceLifecycle, Submittable

SIMULATION_JSON = "simulation.json"
SIM_FILE_HDF5_GZ = "simulation.hdf5.gz"
MODESOLVER_API = "tidy3d/modesolver/py"
MODESOLVER_JSON = "mode_solver.json"
MODESOLVER_HDF5 = "mode_solver.hdf5"
MODESOLVER_GZ = "mode_solver.hdf5.gz"

MODESOLVER_LOG = "output/result.log"
MODESOLVER_RESULT = "output/result.hdf5"
MODESOLVER_RESULT_GZ = "output/mode_solver_data.hdf5.gz"

DEFAULT_NUM_WORKERS = 10
DEFAULT_MAX_RETRIES = 3
DEFAULT_RETRY_DELAY = 10  # in seconds


def run(
    mode_solver: ModeSolver,
    task_name: str = "Untitled",
    mode_solver_name: str = "mode_solver",
    folder_name: str = "Mode Solver",
    results_file: str = "mode_solver.hdf5",
    verbose: bool = True,
    progress_callback_upload: Optional[Callable[[float], None]] = None,
    progress_callback_download: Optional[Callable[[float], None]] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    pay_type: Union[PayType, str] = PayType.AUTO,
) -> ModeSolverData:
    """Submits a :class:`.ModeSolver` to server, starts running, monitors progress, downloads,
    and loads results as a :class:`.ModeSolverData` object.

    Parameters
    ----------
    mode_solver : :class:`.ModeSolver`
        Mode solver to upload to server.
    task_name : str = "Untitled"
        Name of task.
    mode_solver_name: str = "mode_solver"
        The name of the mode solver to create the in task.
    folder_name : str = "Mode Solver"
        Name of folder to store task on web UI.
    results_file : str = "mode_solver.hdf5"
        Path to download results file (.hdf5).
    verbose : bool = True
        If ``True``, will print status, otherwise, will run silently.
    progress_callback_upload : Callable[[float], None] = None
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    progress_callback_download : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.
    reduce_simulation : Literal["auto", True, False] = "auto"
        Restrict simulation to mode solver region. If "auto", then simulation is automatically
        restricted if it contains custom mediums.
    pay_type: Union[PayType, str] = PayType.AUTO
        Which method to pay the simulation.
    Returns
    -------
    :class:`.ModeSolverData`
        Mode solver data with the calculated results.
    """
    log_level = "DEBUG" if verbose else "INFO"
    if verbose:
        console = get_logging_console()

    if reduce_simulation == "auto":
        sim_mediums = mode_solver.simulation.scene.mediums
        contains_custom = any(isinstance(med, AbstractCustomMedium) for med in sim_mediums)
        reduce_simulation = contains_custom

        if reduce_simulation:
            log.warning(
                "The associated 'Simulation' object contains custom mediums. It will be "
                "automatically restricted to the mode solver plane to reduce data for uploading. "
                "To force uploading the original 'Simulation' object use 'reduce_simulation=False'."
                " Setting 'reduce_simulation=True' will force simulation reduction in all cases and"
                " silence this warning."
            )

    if reduce_simulation:
        mode_solver = mode_solver.reduced_simulation_copy

    task = ModeSolverTask.create(mode_solver, task_name, mode_solver_name, folder_name)
    if verbose:
        console.log(
            f"Mode solver created with task_id='{task.task_id}', solver_id='{task.solver_id}'."
        )
    task.upload(verbose=verbose, progress_callback=progress_callback_upload)
    task.submit(pay_type=pay_type)

    # Wait for task to finish
    prev_status = "draft"
    status = task.status
    while status not in ("success", "error", "diverged", "deleted"):
        if status != prev_status:
            log.log(log_level, f"Mode solver status: {status}")
            if verbose:
                console.log(f"Mode solver status: {status}")
            prev_status = status
        time.sleep(0.5)
        status = task.get_info().status

    if status == "error":
        raise WebError("Error running mode solver.")

    log.log(log_level, f"Mode solver status: {status}")
    if verbose:
        console.log(f"Mode solver status: {status}")

    if status != "success":
        # Our cache discards None, so the user is able to re-run
        return None

    return task.get_result(
        to_file=results_file, verbose=verbose, progress_callback=progress_callback_download
    )


def run_batch(
    mode_solvers: list[ModeSolver],
    task_name: str = "BatchModeSolver",
    folder_name: str = "BatchModeSolvers",
    results_files: Optional[list[str]] = None,
    verbose: bool = True,
    max_workers: int = DEFAULT_NUM_WORKERS,
    max_retries: int = DEFAULT_MAX_RETRIES,
    retry_delay: float = DEFAULT_RETRY_DELAY,
    progress_callback_upload: Optional[Callable[[float], None]] = None,
    progress_callback_download: Optional[Callable[[float], None]] = None,
) -> list[ModeSolverData]:
    """
    Submits a batch of ModeSolver to the server concurrently, manages progress, and retrieves results.

    Parameters
    ----------
    mode_solvers : List[ModeSolver]
        List of mode solvers to be submitted to the server.
    task_name : str
        Base name for tasks. Each task in the batch will have a unique index appended to this base name.
    folder_name : str
        Name of the folder where tasks are stored on the server's web UI.
    results_files : List[str], optional
        List of file paths where the results for each ModeSolver should be downloaded. If None, a default path based on the folder name and index is used.
    verbose : bool
        If True, displays a progress bar. If False, runs silently.
    max_workers : int
        Maximum number of concurrent workers to use for processing the batch of simulations.
    max_retries : int
        Maximum number of retries for each simulation in case of failure before giving up.
    retry_delay : int
        Delay in seconds between retries when a simulation fails.
    progress_callback_upload : Callable[[float], None], optional
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    progress_callback_download : Callable[[float], None], optional
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.


    Returns
    -------
    List[ModeSolverData]
        A list of ModeSolverData objects containing the results from each simulation in the batch. ``None`` is placed in the list for simulations that fail after all retries.
    """
    console = get_logging_console()

    if not all(isinstance(x, ModeSolver) for x in mode_solvers):
        console.log(
            "Validation Error: All items in `mode_solvers` must be instances of `ModeSolver`."
        )
        return []

    num_mode_solvers = len(mode_solvers)

    if results_files is None:
        results_files = [f"mode_solver_batch_results_{i}.hdf5" for i in range(num_mode_solvers)]

    def handle_mode_solver(index, progress, pbar):
        retries = 0
        while retries <= max_retries:
            try:
                result = run(
                    mode_solver=mode_solvers[index],
                    task_name=f"{task_name}_{index}",
                    mode_solver_name=f"mode_solver_batch_{index}",
                    folder_name=folder_name,
                    results_file=results_files[index],
                    verbose=False,
                    progress_callback_upload=progress_callback_upload,
                    progress_callback_download=progress_callback_download,
                )
                if verbose:
                    progress.update(pbar, advance=1)
                return result
            except Exception as e:
                console.log(f"Error in mode solver {index}: {e!s}")
                if retries < max_retries:
                    time.sleep(retry_delay)
                    retries += 1
                else:
                    console.log(f"The {index}-th mode solver failed after {max_retries} tries.")
                    if verbose:
                        progress.update(pbar, advance=1)
                    return None

    if verbose:
        console.log(f"[cyan]Running a batch of [deep_pink4]{num_mode_solvers} mode solvers.\n")

        # Create the common folder before running the parallel computation
        _ = Folder.create(folder_name=folder_name)

        with Progress(console=console) as progress:
            pbar = progress.add_task("Status:", total=num_mode_solvers)
            results = Parallel(n_jobs=max_workers, backend="threading")(
                delayed(handle_mode_solver)(i, progress, pbar) for i in range(num_mode_solvers)
            )
            # Make sure the progress bar is complete
            progress.update(pbar, completed=num_mode_solvers, refresh=True)
            console.log("[green]A batch of `ModeSolver` tasks completed successfully!")
    else:
        results = Parallel(n_jobs=max_workers, backend="threading")(
            delayed(handle_mode_solver)(i, None, None) for i in range(num_mode_solvers)
        )

    return results


class ModeSolverTask(ResourceLifecycle, Submittable, extra=pydantic.Extra.allow):
    """Interface for managing the running of a :class:`.ModeSolver` task on server."""

    task_id: str = pydantic.Field(
        None,
        title="task_id",
        description="Task ID number, set when the task is created, leave as None.",
        alias="refId",
    )

    solver_id: str = pydantic.Field(
        None,
        title="solver",
        description="Solver ID number, set when the task is created, leave as None.",
        alias="id",
    )

    real_flex_unit: float = pydantic.Field(
        None, title="real FlexCredits", description="Billed FlexCredits.", alias="charge"
    )

    created_at: Optional[datetime] = pydantic.Field(
        title="created_at", description="Time at which this task was created.", alias="createdAt"
    )

    status: str = pydantic.Field(
        None,
        title="status",
        description="Mode solver task status.",
    )

    file_type: str = pydantic.Field(
        None,
        title="file_type",
        description="File type used to upload the mode solver.",
        alias="fileType",
    )

    mode_solver: ModeSolver = pydantic.Field(
        None,
        title="mode_solver",
        description="Mode solver being run by this task.",
    )

    @classmethod
    def create(
        cls,
        mode_solver: ModeSolver,
        task_name: str = "Untitled",
        mode_solver_name: str = "mode_solver",
        folder_name: str = "Mode Solver",
    ) -> ModeSolverTask:
        """Create a new mode solver task on the server.

        Parameters
        ----------
        mode_solver: :class".ModeSolver"
            The object that will be uploaded to server in the submitting phase.
        task_name: str = "Untitled"
            The name of the task.
        mode_solver_name: str = "mode_solver"
            The name of the mode solver to create the in task.
        folder_name: str = "Mode Solver"
            The name of the folder to store the task.

        Returns
        -------
        :class:`ModeSolverTask`
            :class:`ModeSolverTask` object containing information about the created task.
        """
        folder = Folder.get(folder_name, create=True)

        mode_solver.validate_pre_upload()
        if isinstance(mode_solver.simulation, Simulation):
            mode_solver.simulation.validate_pre_upload(source_required=False)
        elif isinstance(mode_solver.simulation, EMESimulation):
            # TODO: replace this with native web api support
            raise SetupError(
                "'EMESimulation' is not yet supported in the "
                "remote mode solver web api. Please instead call 'ModeSolver.to_fdtd_mode_solver' "
                "before using the web api; this replaces the 'EMESimulation' with a 'Simulation' "
                "that can be used in the remote mode solver. "
                "Alternatively, you can add a 'ModeSolverMonitor' to the 'EMESimulation' "
                "and use the EME solver web api."
            )
            # mode_solver.simulation.validate_pre_upload()
        else:
            raise SetupError("Simulation type not supported in the remote mode solver web api.")

        response_body = {
            "projectId": folder.folder_id,
            "taskName": task_name,
            "protocolVersion": __version__,
            "modeSolverName": mode_solver_name,
            "fileType": "Gz",
            "source": "Python",
        }

        resp = http.post(MODESOLVER_API, response_body)

        # TODO: actually fix the root cause later.
        # sometimes POST for mode returns None and crashes the log.info call.
        # For now this catches it and raises a better error that we can debug.
        if resp is None:
            raise WebError(
                "'ModeSolver' POST request returned 'None'. If you received this error, please "
                "raise an issue on the Tidy3D front end GitHub repository referencing the following"
                f"response body '{response_body}'."
            )

        log.info(
            "Mode solver created with task_id='%s', solver_id='%s'.", resp["refId"], resp["id"]
        )
        return ModeSolverTask(**resp, mode_solver=mode_solver)

    @classmethod
    def get(
        cls,
        task_id: str,
        solver_id: str,
        to_file: str = "mode_solver.hdf5",
        sim_file: str = "simulation.hdf5",
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> ModeSolverTask:
        """Get mode solver task from the server by id.

        Parameters
        ----------
        task_id: str
            Unique identifier of the task on server.
        solver_id: str
            Unique identifier of the mode solver in the task.
        to_file: str = "mode_solver.hdf5"
            File to store the mode solver downloaded from the task.
        sim_file: str = "simulation.hdf5"
            File to store the simulation downloaded from the task.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        :class:`ModeSolverTask`
            :class:`ModeSolverTask` object containing information about the task.
        """
        resp = http.get(f"{MODESOLVER_API}/{task_id}/{solver_id}")
        task = ModeSolverTask(**resp)
        mode_solver = task.get_modesolver(to_file, sim_file, verbose, progress_callback)
        return task.copy(update={"mode_solver": mode_solver})

    def get_info(self) -> ModeSolverTask:
        """Get the current state of this task on the server.

        Returns
        -------
        :class:`ModeSolverTask`
            :class:`ModeSolverTask` object containing information about the task, without the mode
            solver.
        """
        resp = http.get(f"{MODESOLVER_API}/{self.task_id}/{self.solver_id}")
        return ModeSolverTask(**resp, mode_solver=self.mode_solver)

    def upload(
        self, verbose: bool = True, progress_callback: Optional[Callable[[float], None]] = None
    ) -> None:
        """Upload this task's 'mode_solver' to the server.

        Parameters
        ----------
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while uploading the data.
        """
        mode_solver = self.mode_solver.copy()

        sim = mode_solver.simulation

        # Upload simulation.hdf5.gz for GUI display
        file, file_name = tempfile.mkstemp(".hdf5.gz")
        os.close(file)
        try:
            sim.to_hdf5_gz(file_name)
            upload_file(
                self.task_id,
                file_name,
                SIM_FILE_HDF5_GZ,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        finally:
            os.unlink(file_name)

        # Upload a single HDF5 file with the full data
        file, file_name = tempfile.mkstemp(".hdf5.gz")
        os.close(file)
        try:
            mode_solver.to_hdf5_gz(file_name)
            upload_file(
                self.solver_id,
                file_name,
                MODESOLVER_GZ,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        finally:
            os.unlink(file_name)

    def submit(
        self,
        pay_type: Union[PayType, str] = PayType.AUTO,
    ):
        """Start the execution of this task.

        The mode solver must be uploaded to the server with the :meth:`ModeSolverTask.upload` method
        before this step.
        """
        # convert right before sending to API
        pay_type = PayType(pay_type) if not isinstance(pay_type, PayType) else pay_type

        http.post(
            f"{MODESOLVER_API}/{self.task_id}/{self.solver_id}/run",
            {
                "enableCaching": Env.current.enable_caching,
                "payType": pay_type.value,
            },
        )

    def delete(self):
        """Delete the mode solver and its corresponding task from the server."""
        # Delete mode solver
        http.delete(f"{MODESOLVER_API}/{self.task_id}/{self.solver_id}")
        # Delete parent task
        http.delete(f"tidy3d/tasks/{self.task_id}")

    def abort(self):
        """Abort the mode solver and its corresponding task from the server."""
        return http.put(
            "tidy3d/tasks/abort", json={"taskType": "MODE_SOLVER", "taskId": self.solver_id}
        )

    def get_modesolver(
        self,
        to_file: str = "mode_solver.hdf5",
        sim_file: str = "simulation.hdf5",
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> ModeSolver:
        """Get mode solver associated with this task from the server.

        Parameters
        ----------
        to_file: str = "mode_solver.hdf5"
            File to store the mode solver downloaded from the task.
        sim_file: str = "simulation.hdf5"
            File to store the simulation downloaded from the task, if any.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        :class:`.ModeSolver`
            :class:`.ModeSolver` object associated with this task.
        """
        if self.file_type == "Gz":
            file, file_path = tempfile.mkstemp(".hdf5.gz")
            os.close(file)
            try:
                download_file(
                    self.solver_id,
                    MODESOLVER_GZ,
                    to_file=file_path,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
                mode_solver = ModeSolver.from_hdf5_gz(file_path)
            finally:
                os.unlink(file_path)

        elif self.file_type == "Hdf5":
            file, file_path = tempfile.mkstemp(".hdf5")
            os.close(file)
            try:
                download_file(
                    self.solver_id,
                    MODESOLVER_HDF5,
                    to_file=file_path,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
                mode_solver = ModeSolver.from_hdf5(file_path)
            finally:
                os.unlink(file_path)

        else:
            file, file_path = tempfile.mkstemp(".json")
            os.close(file)
            try:
                download_file(
                    self.solver_id,
                    MODESOLVER_JSON,
                    to_file=file_path,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
                mode_solver_dict = ModeSolver.dict_from_json(file_path)
            finally:
                os.unlink(file_path)

            download_file(
                self.task_id,
                SIMULATION_JSON,
                to_file=sim_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
            mode_solver_dict["simulation"] = Simulation.from_json(sim_file)
            mode_solver = ModeSolver.parse_obj(mode_solver_dict)

        # Store requested mode solver file
        mode_solver.to_file(to_file)

        return mode_solver

    def get_result(
        self,
        to_file: str = "mode_solver_data.hdf5",
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> ModeSolverData:
        """Get mode solver results for this task from the server.

        Parameters
        ----------
        to_file: str = "mode_solver_data.hdf5"
            File to store the mode solver downloaded from the task.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        :class:`.ModeSolverData`
            Mode solver data with the calculated results.
        """

        file = None
        try:
            file = download_gz_file(
                resource_id=self.solver_id,
                remote_filename=MODESOLVER_RESULT_GZ,
                to_file=to_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        except ClientError:
            if verbose:
                console = get_logger_console()
                console.log(f"Unable to download '{MODESOLVER_RESULT_GZ}'.")

        if not file:
            try:
                file = download_file(
                    resource_id=self.solver_id,
                    remote_filename=MODESOLVER_RESULT,
                    to_file=to_file,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
            except Exception as e:
                raise WebError(
                    "Failed to download the simulation data file from the server. "
                    "Please confirm that the task was successfully run."
                ) from e

        data = ModeSolverData.from_hdf5(to_file)
        data = data.copy(
            update={"monitor": self.mode_solver.to_mode_solver_monitor(name=MODE_MONITOR_NAME)}
        )

        self.mode_solver._cached_properties["data_raw"] = data

        # Perform symmetry expansion
        self.mode_solver._cached_properties.pop("data", None)
        return self.mode_solver.data

    def get_log(
        self,
        to_file: str = "mode_solver.log",
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Get execution log for this task from the server.

        Parameters
        ----------
        to_file: str = "mode_solver.log"
            File to store the mode solver downloaded from the task.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        return download_file(
            self.solver_id,
            MODESOLVER_LOG,
            to_file=to_file,
            verbose=verbose,
            progress_callback=progress_callback,
        )



from __future__ import annotations

from typing import Any, Optional

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.log import log

MATPLOTLIB_IMPORTED = True
try:
    from matplotlib.colors import is_color_like
except ImportError:
    is_color_like = None
    MATPLOTLIB_IMPORTED = False


def is_valid_color(value: str) -> str:
    if not MATPLOTLIB_IMPORTED:
        log.warning(
            "matplotlib was not successfully imported, but is required "
            "to validate colors in the VisualizationSpec. The specified colors "
            "have not been validated."
        )
    else:
        if is_color_like is not None and not is_color_like(value):
            raise ValueError(f"{value} is not a valid plotting color")

    return value


class VisualizationSpec(Tidy3dBaseModel):
    """Defines specification for visualization when used with plotting functions."""

    facecolor: str = pd.Field(
        "",
        title="Face color",
        description="Color applied to the faces in visualization.",
    )

    edgecolor: Optional[str] = pd.Field(
        "",
        title="Edge color",
        description="Color applied to the edges in visualization.",
    )

    alpha: Optional[pd.confloat(ge=0.0, le=1.0)] = pd.Field(
        1.0,
        title="Opacity",
        description="Opacity/alpha value in plotting between 0 and 1.",
    )

    @pd.validator("facecolor", always=True)
    def validate_color(value: str) -> str:
        return is_valid_color(value)

    @pd.validator("edgecolor", always=True)
    def validate_and_copy_color(value: str, values: dict[str, Any]) -> str:
        if (value == "") and "facecolor" in values:
            return is_valid_color(values["facecolor"])

        return is_valid_color(value)
